import json
import logging
import os
import subprocess
import threading
import time
from collections import OrderedDict
from configparser import SectionProxy
from tempfile import mkstemp
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple

from redun.config import create_config_section
from redun.executors import aws_utils
from redun.executors.base import Executor, register_executor
from redun.executors.code_packaging import package_code, parse_code_package_config
from redun.executors.command import get_oneshot_command, get_script_task_command
from redun.executors.scratch import (
    SCRATCH_OUTPUT,
    SCRATCH_STATUS,
    get_job_scratch_dir,
    get_job_scratch_file,
    parse_job_error,
    parse_job_result,
)
from redun.file import File
from redun.scheduler import Job, Scheduler
from redun.scripting import get_task_command
from redun.task import Task

SUCCEEDED = "SUCCEEDED"
FAILED = "FAILED"


class DockerError(Exception):
    pass


def get_docker_executor_config(config: SectionProxy) -> SectionProxy:
    """
    Returns a config for DockerExecutor.
    """
    keys = ["image", "scratch", "job_monitor_interval", "vcpus", "gpus", "memory"]
    return create_config_section({key: config[key] for key in keys if key in config})


def run_docker(
    command: List[str],
    image: str,
    volumes: Iterable[Tuple[str, str]] = [],
    interactive: bool = True,
    cleanup: bool = False,
    memory: int = 4,
    vcpus: int = 1,
    gpus: int = 0,
    shared_memory: Optional[int] = None,
    include_aws_env: bool = False,
) -> str:
    """
    Run a Docker container locally.

    Parameters
    ----------
    command : List[str]
        A shell command to run within the docker container (e.g. ["ls" "-la"]).
    image : str
        A Docker image.
    volumes : Iterable[Tuple[str, srt]]
        A list of ('host', 'container') path pairs for volume mounting.
    interactive : bool
        If True, the Docker container is run in interactive mode.
    cleanup : bool
        If True, remove the container after execution.
    memory : int
        Number of GB of memory to reserve for the container.
    vcpus : int
        Number of CPUs to reserve for the container.
    gpus : int
        Number of GPUs to reserve for the container.
    shared_memory : Optional[int]
        Number of GB of shared memory to reserve for the container.
    include_aws_env : bool
        If True, forward AWS environment variables to the container.
    """
    # Add AWS credentials to environment for docker command.
    common_args = []
    if cleanup:
        common_args.append("--rm")

    env = dict(os.environ)
    if include_aws_env and not aws_utils.is_ec2_instance():
        # Forward AWS environment args to docker container.
        env.update(aws_utils.get_aws_env_vars())
        common_args.extend(
            [
                "-e",
                "AWS_ACCESS_KEY_ID",
                "-e",
                "AWS_SECRET_ACCESS_KEY",
                "-e",
                "AWS_SESSION_TOKEN",
                "-e",
                "AWS_DEFAULT_REGION",
            ]
        )

    # Volume mounting args.
    for host, container in volumes:
        common_args.extend(["-v", f"{host}:{container}"])

    common_args.extend([f"--memory={memory}g", f"--cpus={vcpus}"])
    if gpus:
        # We can't easily assign a single gpu so we make all available if any GPUs are required.
        common_args.extend(["--gpus", "all"])
    if shared_memory is not None:
        common_args.append(f"--shm-size={shared_memory}g")

    common_args.append(image)
    common_args.extend(command)

    if interactive:
        # Adding this flag is necessary to prevent docker from hijacking the terminal and modifying
        # the tty settings. One of the modifications it makes when it hijacks the terminal is to
        # change the behavior of line endings which means output(like from logging) will be
        # malformed until the docker container exits and the hijacked connection is closed which
        # resets the tty settings.
        env["NORAW"] = "true"

        # Run Docker interactively.
        fd, cidfile = mkstemp()
        os.close(fd)
        os.remove(cidfile)

        docker_command = ["docker", "run", "-it", "--cidfile", cidfile] + common_args
        try:
            subprocess.check_call(docker_command, env=env)
        except subprocess.CalledProcessError as error:
            raise DockerError(error)
        with open(cidfile) as infile:
            container_id = infile.read().strip()
        os.remove(cidfile)
    else:
        # Run Docker in the background.
        docker_command = ["docker", "run", "-d"] + common_args
        try:
            container_id = subprocess.check_output(docker_command, env=env).strip().decode("utf8")
        except subprocess.CalledProcessError as error:
            raise DockerError(error)

    return container_id


def get_docker_job_options(job_options: dict, scratch_path: str) -> dict:
    """
    Returns Docker-specific job options from general job options.

    Adds the scratch_path as a volume mount.
    """
    keys = [
        "vcpus",
        "memory",
        "gpus",
        "shared_memory",
        "volumes",
        "interactive",
        "include_aws_env",
    ]
    options = {key: job_options[key] for key in keys if key in job_options}
    options["volumes"] = options.get("volumes", []) + [(scratch_path, scratch_path)]
    return options


def submit_task(
    image: str,
    scratch_prefix: str,
    job: Job,
    a_task: Task,
    args: Tuple = (),
    kwargs: Dict[str, Any] = {},
    job_options: dict = {},
    code_file: Optional[File] = None,
    include_aws_env: bool = False,
) -> Dict[str, Any]:
    """
    Submit a redun Task to Docker.
    """
    command = get_oneshot_command(
        scratch_prefix,
        job,
        a_task,
        args,
        kwargs,
        job_options=job_options,
        code_file=code_file,
    )
    container_id = run_docker(
        command,
        image=image,
        **get_docker_job_options(job_options, scratch_prefix),
    )
    return {"jobId": container_id, "redun_job_id": job.id}


def submit_command(
    image: str,
    scratch_prefix: str,
    job: Job,
    command: str,
    job_options: dict = {},
    include_aws_env: bool = False,
) -> dict:
    """
    Submit a shell command to Docker.
    """
    shell_command = get_script_task_command(scratch_prefix, job, command)
    container_id = run_docker(
        shell_command,
        image=image,
        **get_docker_job_options(job_options, scratch_prefix),
    )
    return {"jobId": container_id, "redun_job_id": job.id}


def iter_job_status(scratch_prefix: str, job_id2job: Dict[str, "Job"]) -> Iterator[dict]:
    """
    Returns local Docker jobs grouped by their status.
    """
    running_containers = subprocess.check_output(["docker", "ps", "--no-trunc"]).decode("utf8")

    for job_id, redun_job in job_id2job.items():
        if job_id not in running_containers:
            # Job is done running.
            status_file = File(get_job_scratch_file(scratch_prefix, redun_job, SCRATCH_STATUS))
            output_file = File(get_job_scratch_file(scratch_prefix, redun_job, SCRATCH_OUTPUT))

            # Get docker logs and remove container.
            logs = subprocess.check_output(["docker", "logs", job_id]).decode("utf8")
            logs += "Removing container...\n"
            logs += subprocess.check_output(["docker", "rm", job_id]).decode("utf8")

            # TODO: Simplify whether status file is always used or not.
            if status_file.exists():
                succeeded = status_file.read().strip() == "ok"
            else:
                succeeded = output_file.exists()

            status = SUCCEEDED if succeeded else FAILED
            yield {"jobId": job_id, "status": status, "logs": logs}


@register_executor("docker")
class DockerExecutor(Executor):
    """
    A redun Executor for running jobs on local Docker containers.
    """

    def __init__(
        self,
        name: str,
        scheduler: Optional[Scheduler] = None,
        config: Optional[SectionProxy] = None,
    ):
        super().__init__(name, scheduler=scheduler)
        if config is None:
            raise ValueError("DockerExecutor requires config.")
        self._config = config

        # Required config.
        self._image = config["image"]
        self._scratch_prefix_rel = config["scratch"]
        self._scratch_prefix_abs: Optional[str] = None
        self._interval = config.getfloat("job_monitor_interval", fallback=0.2)

        # Optional config.
        self._code_package = parse_code_package_config(config)
        self._code_file: Optional[File] = None

        # Default task options.
        self._default_job_options: Dict[str, Any] = {
            "vcpus": config.getint("vcpus", fallback=1),
            "gpus": config.getint("gpus", fallback=0),
            "memory": config.getint("memory", fallback=4),
            "shared_memory": config.getint("shared_memory", fallback=None),
            "interactive": config.getboolean("interactive", fallback=False),
            "include_aws_env": config.getboolean("include_aws_env", fallback=True),
        }

        self._is_running = False
        self._pending_jobs: Dict[str, "Job"] = OrderedDict()
        self._thread: Optional[threading.Thread] = None

    def set_scheduler(self, scheduler: "Scheduler") -> None:
        super().set_scheduler(scheduler)
        # Perform additional config parsing now that we know our Scheduler.
        self._default_job_options["volumes"] = self._parse_volumes(
            self._config.get("volumes", fallback="[]")
        )

    def _parse_volumes(self, volumes_str: str) -> List[List[str]]:
        """
        Parses Docker volume mounts from JSON string.
        """
        try:
            volumes = json.loads(volumes_str)
        except json.decoder.JSONDecodeError:
            raise ValueError("Invalid 'volumes' syntax. Expected JSON list of path pairs.")

        # Convert host_paths to absolute paths if they are relative.
        # Relative paths are assumed to be relative to the config dir.
        assert self._scheduler
        configdir = self._scheduler.config.configdir
        return [
            [os.path.abspath(os.path.join(configdir, host_path)), container_path]
            for host_path, container_path in volumes
        ]

    @property
    def _scratch_prefix(self) -> str:
        if not self._scratch_prefix_abs:
            if os.path.isabs(self._scratch_prefix_rel):
                self._scratch_prefix_abs = self._scratch_prefix_rel
            else:
                assert self._scheduler
                base_dir = os.path.abspath(self._scheduler.config.configdir)
                self._scratch_prefix_abs = os.path.normpath(
                    os.path.join(base_dir, self._scratch_prefix_rel)
                )
        assert self._scratch_prefix_abs
        return self._scratch_prefix_abs

    def _start(self) -> None:
        """
        Start monitoring thread.
        """
        os.makedirs(self._scratch_prefix, exist_ok=True)

        if not self._is_running:
            self._is_running = True
            self._thread = threading.Thread(target=self._monitor, daemon=False)
            self._thread.start()

    def stop(self) -> None:
        """
        Stop Executor and monitoring thread.
        """
        self._is_running = False

        # Stop monitor thread.
        if (
            self._thread
            and self._thread.is_alive()
            and threading.get_ident() != self._thread.ident
        ):
            self._thread.join()

    def _monitor(self) -> None:
        """
        Thread for monitoring local Docker containers.
        """
        assert self._scheduler

        try:
            while self._is_running and self._pending_jobs:
                # Copy pending_jobs since it can change due to new submissions.
                jobs = iter_job_status(self._scratch_prefix, dict(self._pending_jobs))
                for job in jobs:
                    self._process_job_status(job)
                time.sleep(self._interval)

        except Exception as error:
            # Since we run this is method at the top-level of a thread, we
            # need to catch all exceptions so we can properly report them to
            # the scheduler.
            self._scheduler.reject_job(None, error)

        self.log("Shutting down executor...", level=logging.DEBUG)
        self.stop()

    def _process_job_status(self, job: dict) -> None:
        """
        Process container job statuses.
        """
        assert self._scheduler

        if job["status"] == SUCCEEDED:
            # Assume a recently completed job has valid results.
            redun_job = self._pending_jobs.pop(job["jobId"])
            result, exists = parse_job_result(self._scratch_prefix, redun_job)
            if exists:
                self._scheduler.done_job(redun_job, result)
            else:
                # This can happen if job ended in an inconsistent state.
                self._scheduler.reject_job(
                    redun_job,
                    FileNotFoundError(
                        get_job_scratch_file(self._scratch_prefix, redun_job, SCRATCH_OUTPUT)
                    ),
                )
        elif job["status"] == FAILED:
            redun_job = self._pending_jobs.pop(job["jobId"])
            error, error_traceback = parse_job_error(self._scratch_prefix, redun_job)
            error_traceback.logs = [line + "\n" for line in job["logs"].split("\n")]
            self._scheduler.reject_job(redun_job, error, error_traceback=error_traceback)

    def _submit(self, job: Job) -> None:
        """
        Submit Job to executor.
        """
        assert self._scheduler
        assert job.args
        args, kwargs = job.args

        # Package code if necessary and we have not already done so. If code_package is False,
        # then we can skip this step. Additionally, if we have already packaged and set code_file,
        # then we do not need to repackage.
        if self._code_package is not False and self._code_file is None:
            code_package = self._code_package or {}
            assert isinstance(code_package, dict)
            self._code_file = package_code(self._scratch_prefix, code_package)

        job_options: dict = {
            **self._default_job_options,
            **job.get_options(),
        }
        image: str = job_options.pop("image", self._image)

        # Submit a new Batch job.
        try:
            if not job.task.script:
                docker_resp = submit_task(
                    image,
                    self._scratch_prefix,
                    job,
                    job.task,
                    args=args,
                    kwargs=kwargs,
                    job_options=job_options,
                    code_file=self._code_file,
                )
            else:
                command = get_task_command(job.task, args, kwargs)
                docker_resp = submit_command(
                    image,
                    self._scratch_prefix,
                    job,
                    command,
                    job_options=job_options,
                )
        except DockerError:
            error, error_traceback = parse_job_error(self._scratch_prefix, job)
            self._scheduler.reject_job(job, error, error_traceback=error_traceback)
            return

        job_dir = get_job_scratch_dir(self._scratch_prefix, job)
        self.log(
            "submit redun job {redun_job} as Docker container {container_id}:\n"
            "  container_id = {container_id}\n"
            "  scratch_path = {job_dir}\n".format(
                redun_job=job.id,
                container_id=docker_resp["jobId"],
                job_dir=job_dir,
            )
        )
        self._pending_jobs[docker_resp["jobId"]] = job
        self._start()

    def submit(self, job: Job) -> None:
        """
        Submit Job to executor.
        """
        return self._submit(job)

    def submit_script(self, job: Job) -> None:
        """
        Submit Job for script task to executor.
        """
        return self._submit(job)

    def scratch_root(self) -> str:
        return self._scratch_prefix



import io
import json
import os
import pickle
import sys
import time
from socket import socket
from typing import Any, List, cast
from unittest.mock import Mock, patch

import boto3
import botocore
import pytest
from freezegun import freeze_time
from moto import mock_logs
from sqlalchemy.orm import Session

import redun
from redun import File, Scheduler, task
from redun.backends.db import CallNode, Execution, Job, RedunBackendDb, Value
from redun.backends.db.query import find_file
from redun.cli import (
    REDUN_CONFIG_ENV,
    CallGraphQuery,
    RedunClient,
    RedunClientError,
    check_version,
    find_config_dir,
    get_config_dir,
    import_script,
    is_port_in_use,
    parse_func_path,
    setup_scheduler,
)
from redun.executors.aws_batch import BATCH_LOG_GROUP
from redun.executors.code_packaging import create_tar
from redun.job_array import AWS_ARRAY_VAR
from redun.scheduler import SchedulerError, Traceback
from redun.scheduler_config import get_abs_db_uri
from redun.tags import ANY_VALUE
from redun.tests.utils import assert_match_lines, use_tempdir
from redun.utils import get_import_paths, pickle_dump
from redun.value import get_type_registry


def test_check_version() -> None:
    """
    check_version() should handle all specification formats.
    """
    # Test all comparisons.
    assert check_version("1.0", "1.0")
    assert check_version("1.0", "==1.0")
    assert check_version("1.0", ">=1.0")
    assert check_version("1.0", ">=0.9")
    assert check_version("1.0", ">0.9")
    assert check_version("1.0", "<=1.9")
    assert check_version("1.0", "<1.9")

    # Test all negative comparisons.
    assert not check_version("1.1", "1.0")
    assert not check_version("1.1", "==1.0")
    assert not check_version("0.5", ">=1.0")
    assert not check_version("0.9", ">0.9")
    assert not check_version("2.0", "<=1.9")
    assert not check_version("1.9", "<1.9")

    # Test differences in version parts.
    assert check_version("1", "<2.0")
    assert check_version("1.0", "<2")
    assert check_version("1.2.0", ">1.1")

    # Test double digit versions.
    assert check_version("10", ">1")

    # Test multiple specifications.
    assert check_version("1.0", ">0.9,<2")
    assert not check_version("2.0", ">0.9,<2")

    with pytest.raises(ValueError):
        check_version("boom", "0.9")

    with pytest.raises(ValueError):
        check_version("1.0", "boom")


def test_db_uri() -> None:
    """
    db_uri should be relative to the config_dir.
    """
    cwd = "/home/user/code/a/b"
    config_dir = "../../.redun"

    assert (
        get_abs_db_uri("sqlite:///redun.db", config_dir, cwd)
        == "sqlite:////home/user/code/.redun/redun.db"
    )
    assert get_abs_db_uri("sqlite:////a/b/redun.db", config_dir, cwd) == "sqlite:////a/b/redun.db"
    assert (
        get_abs_db_uri("sqlite://host/a/b/redun.db", config_dir, cwd)
        == "sqlite://host/a/b/redun.db"
    )


@use_tempdir
def test_get_config_dir() -> None:
    """
    get_config_dir() should respect the precedence.
    """
    # Get the default config dir.
    assert get_config_dir() == ".redun"

    # Get the CLI-specified config dir.
    assert get_config_dir("redun-config") == "redun-config"

    # Search for config dir in current directory.
    os.makedirs(".redun")
    expected_config_dir = os.path.join(os.getcwd(), ".redun")
    assert get_config_dir() == expected_config_dir

    # Search up parent directories for config dir.
    os.makedirs("aaa/bbb/ccc")
    os.chdir("aaa/bbb/ccc")
    assert get_config_dir() == expected_config_dir

    # Use environment variable for config dir.
    environ = {REDUN_CONFIG_ENV: "redun-config-env"}

    with patch("os.environ", new_callable=lambda: environ):
        get_config_dir() == "redun-config_env"


@use_tempdir
def test_find_config_dir() -> None:
    """
    Should search up directories to find config_dir.
    """
    os.makedirs(".redun")
    os.makedirs("aaa/bbb/ccc")

    expected_config_dir = os.path.join(os.getcwd(), ".redun")
    cwd = os.getcwd()
    assert find_config_dir(cwd) == expected_config_dir
    assert find_config_dir(os.path.join(cwd, "aaa/bbb/ccc")) == expected_config_dir
    assert find_config_dir(os.path.dirname(cwd)) is None


@use_tempdir
def test_setup_scheduler() -> None:
    """
    Ensure the setup of a scheduler from scratch.
    """
    setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists(".redun/redun.ini")
    assert os.path.exists(".redun/redun.db")


@use_tempdir
def test_setup_scheduler_find_config_dir() -> None:
    """
    Ensure the setup of a scheduler from scratch with higher-level config dir.
    """
    os.makedirs(".redun")
    os.makedirs("aaa/bbb/ccc")
    os.chdir("aaa/bbb/ccc")

    setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists("../../../.redun/redun.ini")
    assert os.path.exists("../../../.redun/redun.db")


@use_tempdir
def test_setup_scheduler_custom_db_uri() -> None:
    """
    Ensure we can use a custom db URI.
    """
    os.makedirs(".redun")
    with open(".redun/redun.ini", "w") as out:
        out.write(
            """
[backend]
db_uri = sqlite:///redun2.db
"""
        )

    setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists(".redun/redun.ini")
    assert os.path.exists(".redun/redun2.db")
    assert not os.path.exists(".redun/redun.db")


@use_tempdir
def test_setup_scheduler_ommit_db_uri() -> None:
    """
    Ensure we use default DB URI if omitted in config.
    """
    os.makedirs(".redun")
    with open(".redun/redun.ini", "w") as out:
        out.write(
            """
[backend]
"""
        )

    setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists(".redun/redun.ini")
    assert os.path.exists(".redun/redun.db")


@use_tempdir
def test_setup_scheduler_grandfather_db_uri() -> None:
    """
    Ensure we can use a the grandfather default db URI.
    """
    os.makedirs(".redun")
    with open(".redun/redun.ini", "w") as out:
        out.write(
            """
[backend]
db_uri = sqlite:///.redun/redun.db
"""
        )

    setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists(".redun/redun.ini")
    assert os.path.exists(".redun/redun.db")


@use_tempdir
def test_import_script_filename() -> None:
    """
    import_script should be able to import a filename.
    """
    File("workflow.py").write(
        """
value = 10
"""
    )

    module: Any = import_script("workflow.py")
    assert module.value == 10


@use_tempdir
def test_import_script_filename_deep() -> None:
    """
    import_script should be able to import a filename from a deep directory.
    """
    File("mylib/scripts/workflow.py").write(
        """
import utils
value = 10
"""
    )
    File("mylib/scripts/utils.py").write(
        """
value = 20
"""
    )

    module: Any = import_script("mylib/scripts/workflow.py")
    assert module.value == 10

    # imports relative to workflow.py should work.
    assert module.utils.value == 20

    # dirname of script should be added as an import path.
    [path] = get_import_paths()
    assert path.endswith("mylib/scripts")


@use_tempdir
def test_import_script_module() -> None:
    """
    import_script should be able to import a module.
    """
    File("workflow.py").write(
        """
value = 10
"""
    )

    module: Any = import_script("workflow")
    assert module.value == 10


@use_tempdir
def test_import_script_module_deep() -> None:
    """
    import_script should be able to import a deep module.
    """
    File("lib/workflows/workflow.py").write(
        """
import lib.workflows.utils as utils

value = 10
"""
    )
    File("lib/workflows/utils.py").write(
        """
value = 20
"""
    )
    File("lib/__init__.py").touch()
    File("lib/workflows/__init__.py").touch()

    module: Any = import_script("lib.workflows.workflow")
    assert module.value == 10

    # imports within module should work.
    assert module.utils.value == 20


@patch("redun.cli.print")
@use_tempdir
def test_task_help(print_mock) -> None:
    """
    CLI should show help for task arguments.

    Both simple types (e.g. int) and generic types (e.g. List) should work.
    """
    File("workflow.py").write(
        """
from typing import List

from redun import task

@task()
def main(x: int, y: List[int]):
    return x
"""
    )

    client = RedunClient()
    client.stdout = io.StringIO()
    client.execute(["redun", "run", "workflow.py", "main", "help"])

    client.stdout.seek(0)
    output = client.stdout.read()
    assert "--x X" in output
    assert "--y Y" in output


@patch("redun.cli.print")
@use_tempdir
def test_task_help_file_annotation(print_mock) -> None:
    """
    CLI should not break when one of the task's inputs has redun.File type.
    """
    File("workflow.py").write(
        """
from typing import List

from redun import File
from redun import task

@task()
def main(x: File, y: List[File]):
    return x
"""
    )

    client = RedunClient()
    client.stdout = io.StringIO()
    client.execute(["redun", "run", "workflow.py", "main", "help"])

    client.stdout.seek(0)
    output = client.stdout.read()
    assert "--x X" in output
    assert "--y Y" in output


@use_tempdir
@patch.dict(os.environ, {AWS_ARRAY_VAR: "1"}, clear=True)
def test_oneshot_existing_output() -> None:
    """
    Ensure we can reunite with an existing output file, in array mode
    """
    File("workflow.py").write(
        """
from redun import task
@task()
def task1(x: int):
    return x + 10
"""
    )

    # Set up output file list
    output_file = File("output_files")
    with output_file.open("w") as ofile:
        json.dump(["output_file1", "output_file2"], ofile)

    # Simulate pre-existing job output.
    actual_output = File("output_file1")
    with actual_output.open("wb") as afile:
        pickle_dump("pre computed answer", afile)

    # Run oneshot job.
    client = RedunClient()
    client.execute(
        [
            "redun",
            "oneshot",
            "--array-job",
            "--output",
            output_file.path,
            "--error",
            output_file.path,
            "workflow.py",
            "task1",
            "--x",
            "10",
        ]
    )

    # Check that no compute was actually run
    assert pickle.loads(cast(bytes, actual_output.read("rb"))) == "pre computed answer"


@use_tempdir
def test_oneshot_existing_output_nocache() -> None:
    """
    Ensure we do not reunite with an existing output file if caching is off.
    """
    File("workflow.py").write(
        """
from redun import task
@task()
def task1(x: int):
    return x + 10
"""
    )

    # Simulate pre-existing job output.
    output_file = File("existing.output")
    with output_file.open("wb") as ofile:
        pickle_dump("pre computed answer", ofile)

    # Run oneshot job.
    client = RedunClient()
    client.execute(
        [
            "redun",
            "oneshot",
            "--output",
            output_file.path,
            "--no-cache",
            "workflow.py",
            "task1",
            "--x",
            "10",
        ]
    )

    # Check that task was run
    assert pickle.loads(cast(bytes, output_file.read("rb"))) == 20


@patch("argparse.ArgumentParser.exit")
@patch("redun.cli.print")
@use_tempdir
def test_task_parse_arg(print_mock, arg_exit_mock) -> None:
    """
    CLI should be able to parse task arguments.
    """
    File("workflow.py").write(
        """
from enum import Enum
from typing import Callable, List, Optional
from redun import task

@task()
def task1(x: str, y):
    return (x, y)

@task()
def main(x: int = 10, y: bool = True):
    return (x, y)

class Color(Enum):
    red = 1
    green = 2
    blue = 3

@task()
def multiply(numbers: List[int], m: int) -> List[int]:
    return [n * m for n in numbers]

@task()
def apply_not(values: List[bool]) -> List[bool]:
    return [not v for v in values]

@task()
def color_to_string(colors: List[Color]) -> List[str]:
    return [str(c) for c in colors]


@task()
def optional_arg(x: Optional[int] = None) -> int:
    if not x:
        return -1
    return 10 * x

def hello(name: str) -> str:
    return f"Hello, {name}!"

@task()
def call_hello(func: Callable[[str], str]) -> str:
    # We should be able to pass functions for a Callable that specifies its
    # argument and result types.
    return func("Alice")

@task()
def call_hello2(func: Callable) -> str:
    # We should be able to pass functions for a Callable that does not specify its
    # argument or result types.
    return func("Alice")
"""
    )

    def arg_exit(status=0, message=None):
        raise ValueError(message)

    arg_exit_mock.side_effect = arg_exit
    client = RedunClient()

    # Use default argument.
    assert client.execute(["redun", "run", "workflow.py", "main"]) == (10, True)

    # Parse argument to int.
    assert client.execute(["redun", "run", "workflow.py", "main", "--x", "20"]) == (
        20,
        True,
    )

    # Parse argument to int and bool.
    assert client.execute(
        ["redun", "run", "workflow.py", "main", "--x", "20", "--y", "false"]
    ) == (20, False)

    # Catch invalid arguments.
    with pytest.raises(ValueError, match="argument --x: invalid int value: 'boom'"):
        assert client.execute(["redun", "run", "workflow.py", "main", "--x", "boom"])

    # Parse argument to int and bool.
    assert client.execute(
        ["redun", "run", "workflow.py", "task1", "--x", "hello", "--y", "there"]
    ) == ("hello", "there")

    # Parse argument as list of ints.
    assert client.execute(
        [
            "redun",
            "run",
            "workflow.py",
            "multiply",
            "--numbers",
            "1",
            "2",
            "3",
            "--m",
            "30",
        ]
    ) == [30, 60, 90]

    # Parse argument as list of bools.
    assert client.execute(
        [
            "redun",
            "run",
            "workflow.py",
            "apply_not",
            "--values",
            "true",
            "False",
            "TRUE",
        ]
    ) == [False, True, False]

    # Parse argument as list of enums.
    assert client.execute(
        [
            "redun",
            "run",
            "workflow.py",
            "color_to_string",
            "--colors",
            "Color.red",
            "blue",
        ]
    ) == ["Color.red", "Color.blue"]

    # Confirm list args are validated against their type.
    with pytest.raises(ValueError, match="argument --colors: invalid Color value"):
        # black is not a color in the enum, should raise an error.
        assert client.execute(
            [
                "redun",
                "run",
                "workflow.py",
                "color_to_string",
                "--colors",
                "Color.red",
                "black",
            ]
        )

    # Confirm optional args are not required by the CLI
    assert client.execute(["redun", "run", "workflow.py", "optional_arg"]) == -1

    # Confirm optional args are typed correctly when supplied in the CLI
    assert client.execute(["redun", "run", "workflow.py", "optional_arg", "--x", "2"]) == 20

    # Parse argument as a function.
    assert (
        client.execute(["redun", "run", "workflow.py", "call_hello", "--func", "workflow.hello"])
        == "Hello, Alice!"
    )
    assert (
        client.execute(["redun", "run", "workflow.py", "call_hello2", "--func", "workflow.hello"])
        == "Hello, Alice!"
    )

    # Detect errors when parsing a function name.
    with pytest.raises(
        ValueError,
        match="argument --func: invalid function value: 'bad_workflow.hello'",
    ):
        client.execute(
            [
                "redun",
                "run",
                "workflow.py",
                "call_hello2",
                "--func",
                "bad_workflow.hello",
            ]
        )

    with pytest.raises(
        ValueError, match="argument --func: invalid function value: 'workflow.bad_func'"
    ):
        client.execute(
            [
                "redun",
                "run",
                "workflow.py",
                "call_hello2",
                "--func",
                "workflow.bad_func",
            ]
        )

    with pytest.raises(ValueError, match="argument --func: invalid function value: 'hello'"):
        # Module ('workflow') must be included for the argument `--func`.
        client.execute(["redun", "run", "workflow.py", "call_hello2", "--func", "hello"])


@use_tempdir
def test_invalid_config_args_raises() -> None:
    """
    Tasks with config_args not matching args/kwargs should raise an error.
    """
    File("workflow.py").write(
        """
from redun.task import task
@task(config_args=["this_is_not_valid"])
def add(n1: int, n2: int):
    return n1 + n2
"""
    )

    client = RedunClient()
    with pytest.raises(ValueError):
        client.execute(["redun", "run", "workflow.py", "add", "--n1", "1", "--n2", "10"])


@patch("argparse.ArgumentParser.exit")
@patch("redun.cli.print")
@use_tempdir
def test_task_parse_config_arg(print_mock, arg_exit_mock) -> None:
    """
    CLI should be able to parse task arguments.
    """
    File("workflow.py").write(
        """
from redun import task
@task(config_args=["memory"])
def main(x: int = 10, memory: int = 4):
    return (x, memory)
"""
    )

    def arg_exit(status=0, message=None):
        raise ValueError(message)

    arg_exit_mock.side_effect = arg_exit
    client = RedunClient()

    # Use default config arg.
    assert client.execute(["redun", "run", "workflow.py", "main"]) == (10, 4)

    # Parse config arg to int.
    assert client.execute(
        ["redun", "run", "workflow.py", "main", "--x", "20", "--memory", "5"]
    ) == (20, 5)

    # Different config arg does not force re-execution.
    assert client.execute(
        ["redun", "run", "workflow.py", "main", "--x", "20", "--memory", "10"]
    ) == (20, 5)

    # Catch invalid config arg.
    with pytest.raises(ValueError, match="argument --memory: invalid int value: 'boom'"):
        assert client.execute(
            ["redun", "run", "workflow.py", "main", "--x", "10", "--memory", "boom"]
        )


def test_parse_func_path() -> None:
    """
    `parse_func_path()` should be able to parse the pattern 'file_or_module::function_name'.
    """
    assert parse_func_path("workflow.py::setup_scheduler") == (
        "workflow.py",
        "setup_scheduler",
    )
    assert parse_func_path("workflow::setup_scheduler") == (
        "workflow",
        "setup_scheduler",
    )


@use_tempdir
@patch("argparse.ArgumentParser.exit")
def test_setup_func(arg_exit_mock: Mock) -> None:
    """
    RedunClient should use custom scheduler setup functions.
    """
    File("workflow.py").write(
        """\
from redun import task, Scheduler
from redun.executors.local import LocalExecutor

redun_namespace = 'acme'

@task()
def main(x: int = 10):
    return x + 1

def setup_scheduler(config, name: str = 'exec2', extra: int = 1):
    scheduler = Scheduler(config=config)
    scheduler.extra = extra
    scheduler.add_executor(LocalExecutor(name))
    return scheduler
"""
    )

    File(".redun/redun.ini").write(
        """\
[scheduler]
setup_scheduler = workflow::setup_scheduler
"""
    )

    def arg_exit(*args, **kwargs):
        raise ValueError()

    arg_exit_mock.side_effect = arg_exit
    client = RedunClient()

    # Ensure workflow can run.
    assert client.execute(["redun", "run", "workflow.py", "main"]) == 11

    # An additional executor should have been configured.
    assert client.scheduler
    assert "exec2" in client.scheduler.executors

    # Ensure pass arguments to setip func.
    client = RedunClient()
    assert (
        client.execute(
            [
                "redun",
                "--setup",
                "name=exec3",
                "--setup",
                "extra=100",
                "run",
                "workflow.py",
                "main",
            ]
        )
        == 11
    )

    # An additional executor should have been configured.
    assert client.scheduler
    assert "exec3" in client.scheduler.executors
    assert client.scheduler.extra == 100  # type: ignore

    # Ensure pass arguments to setip func.
    with pytest.raises(ValueError):
        client = RedunClient()
        client.execute(["redun", "--setup", "bad=boom", "run", "workflow.py", "main"])


@use_tempdir
def test_find_file(scheduler: Scheduler, session: Session) -> None:
    @task()
    def task1():
        file = File("out.txt")
        file.write("hello")
        return file

    scheduler.run(task1())

    row = find_file(session, "out.txt")
    assert row
    file, job, kind = row
    assert file.path == "out.txt"
    assert kind == "result"


@use_tempdir
def test_find_file_as_subvalue(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to find Files that are subvalues.
    """

    @task()
    def task1():
        file = File("out.txt")
        file.write("hello")
        # Return file as a subvalue of a list.
        return [file, 10]

    scheduler.run(task1())

    result = find_file(session, "out.txt")
    assert result
    file, job, kind = result
    assert file.path == "out.txt"
    assert kind == "result"


@use_tempdir
def test_find_file_arg(scheduler: Scheduler, session: Session) -> None:
    """
    Find a File given as an Argument to a Task.
    """

    @task()
    def task1(file):
        return 10

    @task()
    def main():
        file = File("out.txt")
        file.write("hello")
        return task1(file)

    scheduler.run(main())

    result = find_file(session, "out.txt")
    assert result
    file, job, kind = result
    assert file.path == "out.txt"
    assert kind == "arg"


@use_tempdir
def test_find_file_arg_subvalue(scheduler: Scheduler, session: Session) -> None:
    """
    Find a File given as a Subvalue in an Argument to a Task.
    """

    @task()
    def task1(file_and_int):
        return 10

    @task()
    def main():
        file = File("out.txt")
        file.write("hello")
        return task1([file, 10])

    scheduler.run(main())

    result = find_file(session, "out.txt")
    assert result
    file, job, kind = result
    assert file.path == "out.txt"
    assert kind == "arg"


@use_tempdir
def test_oneshot_code() -> None:
    """
    Ensure we can run a task from a code package.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def task1(x: int):
    return x + 1
"""
    )

    create_tar("code.tar.gz", ["workflow.py"])

    # Remove the original file to force the use of the code package.
    file.remove()

    client = RedunClient()
    client.execute(
        [
            "redun",
            "oneshot",
            "--code",
            "code.tar.gz",
            "--output",
            "out",
            "workflow.py",
            "task1",
            "--x",
            "10",
        ]
    )

    out = File("out")
    assert pickle.loads(cast(bytes, out.read("rb"))) == 11


@use_tempdir
def test_oneshot_shortcircuit() -> None:
    """
    Ensures oneshot can use existing output.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def task_hi(x: int):
    return "Hello" * x
"""
    )

    output = File("output")
    with output.open("wb") as ofile:
        pickle_dump("Hello" * 3, ofile)

    # Should return saved output.
    client = RedunClient()
    client.execute(
        ["redun", "oneshot", "--output", "output", "workflow.py", "task_hi", "--x", "2"]
    )

    assert pickle.loads(cast(bytes, output.read("rb"))) == "Hello" * 3

    # Should recalculate with no-cache.
    client = RedunClient()
    client.execute(
        [
            "redun",
            "oneshot",
            "--no-cache",
            "--output",
            "output",
            "workflow.py",
            "task_hi",
            "--x",
            "2",
        ]
    )

    assert pickle.loads(cast(bytes, output.read("rb"))) == "Hello" * 2


@use_tempdir
def test_oneshot_import() -> None:
    """
    Ensure we can oneshot a task using import path.
    """
    file = File("path/to/workflow.py")
    file.write(
        """
from redun import task

@task()
def task1(x: int):
    return x + 1
"""
    )

    client = RedunClient()
    client.stdout = io.StringIO()
    client.execute(
        [
            "redun",
            "oneshot",
            "--import-path",
            "path/to",
            "--output",
            "out",
            "workflow",
            "task1",
            "--x",
            "10",
        ]
    )

    out = File("out")
    assert pickle.loads(cast(bytes, out.read("rb"))) == 11


@use_tempdir
@patch("redun.cli.get_aws_user", side_effect=botocore.exceptions.NoCredentialsError())
@patch(
    "redun.cli.get_simple_aws_user",
    side_effect=botocore.exceptions.NoCredentialsError(),
)
@patch("redun.cli.get_gcp_user", lambda: None)
def test_run_tags(get_simple_aws_user_mock: Mock, get_aws_user_mock: Mock) -> None:
    """
    Run tags should be recorded on the execution.
    """
    os.makedirs("acme")
    os.chdir("acme")

    file = File("workflow.py")
    file.write(
        """
from redun import task

redun_namespace = "acme"

@task()
def task1(x: int):
    return x + 1
"""
    )

    client = RedunClient()
    client.execute(
        [
            "redun",
            "run",
            "--user",
            "alice",
            "--tag",
            "cost=10",
            "--tag",
            'env=["dev", "stg"]',
            "workflow.py",
            "task1",
            "--x",
            "10",
        ]
    )

    scheduler = setup_scheduler()
    assert scheduler.backend
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.session

    execution = backend.session.query(Execution).one()
    tags = backend.get_tags([execution.id])[execution.id]
    assert tags == {
        "env": [["dev", "stg"]],
        "user": ["alice"],
        "project": ["acme"],
        "cost": [10],
        "redun.version": [redun.__version__],
    }


@use_tempdir
def test_oneshot_error() -> None:
    """
    oneshot should produce an error file when a task raises an exception.
    """
    file = File("workflow_error.py")
    file.write(
        """
from redun import task

@task()
def task_error(x: int):
    raise ValueError("boom")
"""
    )

    create_tar("code.tar.gz", ["workflow_error.py"])

    # Remove the original file to force the use of the code package.
    file.remove()

    client = RedunClient()
    with pytest.raises(ValueError):
        client.execute(
            [
                "redun",
                "oneshot",
                "--code",
                "code.tar.gz",
                "--output",
                "out",
                "--error",
                "error",
                "workflow_error.py",
                "task_error",
                "--x",
                "10",
            ]
        )

    error_file = File("error")
    error, error_traceback = pickle.loads(cast(bytes, error_file.read("rb")))

    assert isinstance(error, ValueError)
    assert isinstance(error_traceback, Traceback)


@use_tempdir
def test_oneshot_error_bad_input() -> None:
    """
    oneshot should produce an error file when input file is bad.
    """
    file = File("workflow_error.py")
    file.write(
        """
from redun import task

@task()
def task1(x: int):
    return x
"""
    )

    create_tar("code.tar.gz", ["workflow_error.py"])

    # Remove the original file to force the use of the code package.
    file.remove()

    client = RedunClient()
    with pytest.raises(FileNotFoundError):
        client.execute(
            [
                "redun",
                "oneshot",
                "--code",
                "code.tar.gz",
                "--input",
                "does_not_exist",
                "--output",
                "out",
                "--error",
                "error",
                "workflow_error.py",
                "task1",
            ]
        )

    error_file = File("error")
    error, error_traceback = pickle.loads(cast(bytes, error_file.read("rb")))

    assert isinstance(error, FileNotFoundError)
    assert isinstance(error_traceback, Traceback)


@mock_logs
@freeze_time("2020-01-01 00:00:00", tz_offset=-7)
@patch("redun.executors.aws_batch.aws_describe_jobs")
@patch("redun.cli.aws_describe_jobs")
@use_tempdir
def test_aws_logs(aws_describe_jobs_mock1, aws_describe_jobs_mock2) -> None:
    timestamp = int(time.time() * 1000)
    stream_name = "redun_aws_batch_example-redun-jd/default/6c939514f4054fdfb5ee65acc8aa4b07"
    job_descriptions = [
        {
            "jobId": "123",
            "container": {"logStreamName": stream_name},
            "startedAt": timestamp,
            "createdAt": timestamp,
        }
    ]
    aws_describe_jobs_mock1.return_value = iter(job_descriptions)
    aws_describe_jobs_mock2.return_value = iter(job_descriptions)

    # Setup logs mocks.
    logs_client = boto3.client("logs", region_name="us-west-2")
    logs_client.create_log_group(logGroupName=BATCH_LOG_GROUP)
    logs_client.create_log_stream(logGroupName=BATCH_LOG_GROUP, logStreamName=stream_name)
    resp = logs_client.put_log_events(
        logGroupName=BATCH_LOG_GROUP,
        logStreamName=stream_name,
        logEvents=[
            {"timestamp": timestamp, "message": "A message 1."},
            {"timestamp": timestamp + 1, "message": "A message 2."},
        ],
    )
    resp = logs_client.put_log_events(
        logGroupName=BATCH_LOG_GROUP,
        logStreamName=stream_name,
        logEvents=[
            {"timestamp": timestamp + 2, "message": "A message 3."},
            {"timestamp": timestamp + 3, "message": "A message 4."},
        ],
        sequenceToken=resp["nextSequenceToken"],
    )

    File(".redun/redun.ini").write(
        """
[executors.batch]
type = aws_batch
image = my-image
queue = my-queue
s3_scratch = s3://bucket/scratch/
"""
    )

    client = RedunClient()
    client.execute(["redun", "init"])
    client.stdout = io.StringIO()
    client.execute(
        [
            "redun",
            "aws",
            "logs",
            "--batch-job",
            "123",
        ]
    )

    client.stdout.seek(0)

    lines = client.stdout.readlines()
    assert lines == [
        "# AWS Batch job 123\n",
        "# AWS Log Stream redun_aws_batch_example-redun-jd/default/6c939514f4054fdfb5ee65acc8aa4b07\n",  # noqa:E501
        "2019-12-31 17:00:00  A message 1.\n",
        "2019-12-31 17:00:00.001000  A message 2.\n",
        "2019-12-31 17:00:00.002000  A message 3.\n",
        "2019-12-31 17:00:00.003000  A message 4.\n",
        "\n",
    ]


@use_tempdir
def test_project_tag() -> None:
    """
    Run tags should be recorded on the execution.
    """
    os.makedirs("acme")
    os.chdir("acme")

    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def task1(x: int):
    return x + 1
"""
    )

    client = RedunClient()
    client.execute(
        ["redun", "run", "--project", "my-project", "workflow.py", "task1", "--x", "10"]
    )

    scheduler = setup_scheduler()
    assert scheduler.backend
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.session

    execution = backend.session.query(Execution).one()
    tags = backend.get_tags([execution.id])[execution.id]
    assert tags["project"] == ["my-project"]


@use_tempdir
def test_cloud_user_tags() -> None:
    """
    Run cloud (AWS, GCP) user tags should be recorded on the execution.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def task1(x: int):
    return x + 1
"""
    )

    with patch("redun.cli.get_aws_user", lambda _: "arn:::::alice"), patch(
        "redun.cli.get_simple_aws_user", lambda _: "alice"
    ), patch("redun.cli.get_gcp_user", lambda: "alice@gmail.com"):
        client = RedunClient()
        client.execute(["redun", "run", "workflow.py", "task1", "--x", "10"])

    scheduler = setup_scheduler()
    assert scheduler.backend
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.session

    execution = backend.session.query(Execution).one()
    tags = backend.get_tags([execution.id])[execution.id]
    assert tags["user_aws"] == ["alice"]
    assert tags["user_aws_arn"] == ["arn:::::alice"]
    assert tags["user_gcp"] == ["alice@gmail.com"]


def test_query(scheduler: Scheduler, backend: RedunBackendDb, session: Session) -> None:
    """
    CallGraphQuery should query records correctly.
    """

    @task(tags=[("step", "calc")])
    def task1(x):
        return x + 1

    assert scheduler.run(task1(10), tags=[("project", "acme")]) == 11

    # All entities should be part of query.
    query = CallGraphQuery(session)
    assert len(list(query.all())) == 7
    assert len(list(query.order_by("time").all())) == 7
    assert list(query.count()) == [
        ("Execution", 1),
        ("Job", 1),
        ("CallNode", 1),
        ("Task", 1),
        ("Value", 3),
    ]

    # Select ids.
    ids = list(query.select("id", flat=True))
    assert task1.hash in ids

    # Select type and ids.
    type_ids = list(query.select("type", "id"))
    assert ("Task", task1.hash) in type_ids

    # Filter by type.
    execution = query.filter_types(["Execution"]).one()
    assert isinstance(execution, Execution)
    # ensure the correct type is returned when ordering query is added
    execution_ordered = query.order_by("time").filter_types(["Execution"]).one()
    assert isinstance(execution_ordered, Execution)

    call_node = query.filter_types(["CallNode"]).one()
    assert isinstance(call_node, CallNode)

    # Filter by value type.
    assert dict(query.filter_value_types(["builtins.int"]).count())["Value"] == 2

    # Filter by ids.
    assert set(
        query.filter_ids([execution.id, call_node.call_hash, task1.hash]).select("type", "id")
    ) == {
        ("Execution", execution.id),
        ("CallNode", call_node.call_hash),
        ("Task", task1.hash),
        ("Value", task1.hash),
    }

    # Filter by id prefix.
    assert query.like_id(task1.hash[:8]).filter_types(["Task"]).one().hash == task1.hash

    # Generate an extra execution.
    assert scheduler.run(task1(20)) == 21

    # Filter execution id.
    exec_id = cast(Execution, query.filter_types({"Execution"}).first()).id
    assert list(query.count()) == [
        ("Execution", 2),
        ("Job", 2),
        ("CallNode", 2),
        ("Task", 1),
        ("Value", 5),
    ]
    assert list(query.filter_execution_ids([exec_id]).count()) == [
        ("Execution", 1),
        ("Job", 1),
        ("CallNode", 1),
        ("Task", 1),
        ("Value", 2),
    ]

    # Filter by execution status.
    assert len(list(query.filter_execution_statuses(["DONE"]).all()))

    # Filter by jobs status.
    assert len(list(query.filter_job_statuses(["DONE"]).all())) == 2
    assert len(list(query.filter_job_statuses(["CACHED"]).all())) == 0
    assert len(list(query.filter_job_statuses(["FAILED"]).all())) == 0

    # Filter by tags.
    assert dict(query.filter_tags([("step", "calc")]).count()) == {
        "CallNode": 0,
        "Execution": 0,
        "Job": 2,
        "Task": 1,
        "Value": 1,
    }
    assert dict(query.filter_tags([("project", "acme")]).count()) == {
        "CallNode": 0,
        "Execution": 1,
        "Job": 0,
        "Task": 0,
        "Value": 0,
    }

    # Filter by execution tags.
    assert dict(query.filter_execution_tags([("project", "acme")]).count()) == {
        "CallNode": 1,
        "Execution": 1,
        "Job": 1,
        "Task": 1,
        "Value": 2,
    }

    # Empty query should return nothing.
    assert list(query.empty().all()) == []


def test_filter_task_names(
    scheduler: Scheduler, backend: RedunBackendDb, session: Session
) -> None:
    """
    Should be able to filter by task name.
    """

    @task(namespace="ns")
    def task1(x):
        return x + 1

    @task(namespace="ns")
    def main():
        return task1(10)

    assert scheduler.run(main()) == 11

    # All entities should be part of query.
    query = CallGraphQuery(session)

    assert len(list(query.filter_types(["Job"]).filter_task_names(["task1"]).all())) == 1
    assert len(list(query.filter_types(["Job"]).filter_task_names(["ns.task1"]).all())) == 1
    assert len(list(query.filter_types(["Task"]).filter_task_names(["task1"]).all())) == 1
    assert len(list(query.filter_types(["Job"]).filter_task_names(["task1", "main"]).all())) == 2
    assert len(list(query.filter_types(["Task"]).filter_task_names(["bad_task"]).all())) == 0


def test_value_subqueries(scheduler: Scheduler, backend: RedunBackendDb, session: Session) -> None:
    """
    CallGraphQuery should be able to query Values by Execution.
    """

    @task()
    def task1(x: Any) -> int:
        return 10

    @task()
    def task2(x: Any) -> List[File]:
        return [File("foo"), File("bar")]

    # Single Value argument and result.
    scheduler.run(task1(1))

    # List Value argument.
    scheduler.run(task1([File("alice")]))

    # List Value result.
    scheduler.run(task2(2))

    query = CallGraphQuery(session)

    # Get the three executions.
    exec3, exec2, exec1 = list(query.filter_types(["Execution"]).order_by("time").all())
    assert exec3.call_node.timestamp > exec2.call_node.timestamp > exec1.call_node.timestamp

    # We should be able to fetch the other record types from the execution.
    assert (
        query.filter_types(["Job"]).filter_execution_ids([exec1.id]).one().execution_id == exec1.id
    )
    assert query.filter_types(["Task"]).filter_execution_ids([exec1.id]).one().name == "task1"
    assert (
        query.filter_types(["CallNode"])
        .filter_execution_ids([exec1.id])
        .one()
        .jobs[0]
        .execution_id
        == exec1.id
    )

    # We should be able to fetch the Values from an Execution.
    values = query.filter_types(["Value"]).filter_execution_ids([exec1.id]).all()
    assert {value.value_parsed for value in values} == {1, 10}

    type_registry = get_type_registry()

    def assert_hashes(values1, raw_values2):
        hashes1 = {type_registry.get_hash(value.value_parsed) for value in values1}
        hashes2 = {type_registry.get_hash(value) for value in raw_values2}
        assert hashes1 == hashes2

    # We should get both the list and subvalues.
    values2 = list(query.filter_types(["Value"]).filter_execution_ids([exec2.id]).all())
    assert_hashes(values2, [10, File("alice"), [File("alice")]])

    values3 = list(query.filter_types(["Value"]).filter_execution_ids([exec3.id]).all())
    assert_hashes(values3, [2, File("foo"), File("bar"), [File("foo"), File("bar")]])


def test_query_tags(scheduler: Scheduler, backend: RedunBackendDb, session: Session) -> None:
    """
    CallGraphQuery should query tags correctly.
    """

    @task()
    def task1(x):
        return x + 1

    assert (
        scheduler.run(task1(10), tags=[("project", "acme"), ("user", "alice"), ("env", "prod")])
        == 11
    )
    assert (
        scheduler.run(task1(100), tags=[("project", "skunk"), ("user", "bob"), ("env", "prod")])
        == 101
    )

    # Get Executions for testing.
    query = CallGraphQuery(session)
    execs = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time)
        .all()
    )
    exec1, exec2 = execs

    # Query by key and value.
    assert set(query.filter_tags([("project", "acme")]).all()) == {exec1}
    assert set(query.filter_tags([("env", "prod")]).all()) == {exec1, exec2}

    # Query by only key.
    assert set(query.filter_tags([("project", ANY_VALUE)]).all()) == {exec1, exec2}

    # Query by wrong key.
    assert set(query.filter_tags([("unknown", ANY_VALUE)]).all()) == set()

    # Query by two tags.
    assert set(query.filter_tags([("project", "acme"), ("user", "alice")]).all()) == {exec1}

    # Query values by exec tag.
    values = query.filter_types(["Value"]).filter_execution_tags([("project", "acme")]).all()
    assert {value.value_parsed for value in values} == {10, 11}

    values = query.filter_types(["Value"]).filter_execution_tags([("env", "prod")]).all()
    assert {value.value_parsed for value in values} == {10, 11, 100, 101}

    # Query by two exec tags.
    values = (
        query.filter_types(["Value"])
        .filter_execution_tags([("project", "acme"), ("user", "alice")])
        .all()
    )
    assert {value.value_parsed for value in values} == {10, 11}


def run_command(client: RedunClient, argv: List[str]) -> str:
    """
    Run redun cli command and return output as a string.
    """
    client.stdout = io.StringIO()
    client.execute(argv)
    client.stdout.seek(0)
    return client.stdout.read()


@use_tempdir
def test_cli_log() -> None:
    """
    cli should allow querying the CallGraph.
    """
    file = File("workflow.py")
    file.write(
        """
import time

from redun import task

redun_namespace = 'test'

@task()
def task1(x: int):
    return x + 1

@task()
def main(x: int):
    return task1(x)
"""
    )

    # Run example workflow.
    client = RedunClient()
    client.execute(["redun", "run", "workflow.py", "main", "--x", "10"])
    client.execute(["redun", "run", "workflow.py", "main", "--x", "11"])
    client.execute(["redun", "run", "workflow.py", "main", "--x", "10"])

    # Show just executions.
    lines = run_command(client, ["redun", "log"]).split("\n")
    assert_match_lines(
        [
            r"Recent executions:",
            r"",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 10 \(.*\)",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 11 \(.*\)",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 10 \(.*\)",
            r"",
        ],
        lines,
    )

    # Filter by record type.
    lines = run_command(client, ["redun", "log", "--job"]).split("\n")
    assert_match_lines(
        [
            r"Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.task1\(10\) ",
            r"Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.main\(x=10\) ",
            r"Job ........ \[ DONE \] ....-..-.. ..:..:..:  test.task1\(11\) ",
            r"Job ........ \[ DONE \] ....-..-.. ..:..:..:  test.main\(x=11\) ",
            r"Job ........ \[ DONE \] ....-..-.. ..:..:..:  test.task1\(10\) ",
            r"Job ........ \[ DONE \] ....-..-.. ..:..:..:  test.main\(x=10\) ",
            "",
        ],
        lines,
    )

    lines = run_command(client, ["redun", "log", "--task"]).split("\n")
    assert_match_lines(
        ["Task ........ test.main ", "Task ........ test.task1 ", ""],
        lines,
    )

    lines = run_command(client, ["redun", "log", "--exec"]).split("\n")
    assert_match_lines(
        [
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 10 \(.*\)",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 11 \(.*\)",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 10 \(.*\)",
            "",
        ],
        lines,
    )

    # Filter by exec id.
    lines = run_command(client, ["redun", "log", "--exec"]).split("\n")
    exec_id = lines[0].split()[1]
    lines = run_command(client, ["redun", "log", exec_id]).split("\n")
    assert_match_lines(
        [
            r"Exec .* \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 10 \(.*\)",
            r"Duration: 0:00:..\...",
            r"",
            r"\| JOB STATUS ..../../.. ..:..:..",
            r"\| TASK       PENDING RUNNING  FAILED  CACHED    DONE   TOTAL",
            r"\| ",
            r"\| ALL              0       0       0       2       0       2",
            r"\| test\.main        0       0       0       1       0       1",
            r"\| test\.task1       0       0       0       1       0       1",
            r"",
            r"Jobs: 2 \(DONE: 0, CACHED: 2, FAILED: 0\)",
            r"--------------------------------------------------------------------------------",
            r"Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.main\(x=10\) ",
            r"  Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.task1\(10\) ",
            r"",
        ],
        lines,
    )

    # Filter by most recent execution id.
    lines = run_command(client, ["redun", "log", "-"]).split("\n")
    assert_match_lines(
        [
            r"Exec .* \[ DONE \] ....-..-.. ..:..:..:  run workflow.py main --x 10 \(.*\)",
            r"Duration: 0:00:..\...",
            r"",
            r"\| JOB STATUS ..../../.. ..:..:..",
            r"\| TASK       PENDING RUNNING  FAILED  CACHED    DONE   TOTAL",
            r"\| ",
            r"\| ALL              0       0       0       2       0       2",
            r"\| test\.main        0       0       0       1       0       1",
            r"\| test\.task1       0       0       0       1       0       1",
            r"",
            r"Jobs: 2 \(DONE: 0, CACHED: 2, FAILED: 0\)",
            r"--------------------------------------------------------------------------------",
            r"Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.main\(x=10\) ",
            r"  Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.task1\(10\) ",
            r"",
        ],
        lines,
    )

    # Filter jobs by execution id.
    lines = run_command(client, ["redun", "log", "--job", "--exec-id", exec_id]).split("\n")
    assert_match_lines(
        [
            r"Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.task1\(10\) ",
            r"Job ........ \[CACHED\] ....-..-.. ..:..:..:  test.main\(x=10\) ",
            "",
        ],
        lines,
    )

    # Job detail.
    lines = run_command(client, ["redun", "log", "--job"]).split("\n")
    job_id = lines[0].split()[1]
    lines = run_command(client, ["redun", "log", job_id]).split("\n")
    assert_match_lines(
        [
            r"Job .* \[CACHED\] ....-..-.. ..:..:..:  test.task1\(10\) ",
            r"Traceback: Exec ........ > Job ........ main > Job ........ task1",
            r"",
            r"  CallNode .* test.task1 ",
            r"    Args:   10",
            r"    Result: 11",
            r"",
            r"  Task .* test.task1 ",
            r"",
            r"    def task1\(x: int\):",
            r"        return x \+ 1",
            r"",
            r"",
            r"  Upstream dataflow:",
            r"",
            r"    result = 11",
            r"",
            r"    result <-- <........> task1\(x\)",
            r"      x = <........> 10",
            r"",
            r"    x <-- argument of <........> main\(x\)",
            r"      <-- origin",
            r"",
            r"",
        ],
        lines,
    )


@use_tempdir
@patch("redun.cli.get_aws_user", side_effect=botocore.exceptions.NoCredentialsError())
@patch(
    "redun.cli.get_simple_aws_user",
    side_effect=botocore.exceptions.NoCredentialsError(),
)
@patch("redun.cli.get_gcp_user", lambda: None)
def test_cli_log_tags(get_simple_aws_user_mock: Mock, get_aws_user_mock: Mock) -> None:
    """
    cli should allow querying the CallGraph for tags.
    """
    file = File("workflow.py")
    file.write(
        """
import time

from redun import task

redun_namespace = 'test'

@task(tags=[("step", "calc")])
def task1(x: int) -> int:
    return x + 1

@task()
def main(x: int) -> int:
    return task1(x)
"""
    )

    # Run example workflow.
    client = RedunClient()
    client.execute(
        [
            "redun",
            "run",
            "--user",
            "alice",
            "workflow.py",
            "task1",
            "--x",
            "10",
        ]
    )

    scheduler = setup_scheduler()
    assert scheduler.backend
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.session

    execution = backend.session.query(Execution).one()

    # Query all tags.
    output = run_command(client, ["redun", "tag", "list"])
    assert (
        output == "Tags:\n  project    (1 Execution)\n"
        "  redun.version (1 Execution)\n"
        "  step       (1 Job, 1 Task)\n"
        "  user       (1 Execution)\n"
    )

    # Query specific tag key.
    output = run_command(client, ["redun", "tag", "list", "step"])
    assert "test.task1 (step=calc)" in output

    # Query tag key=value.
    output = run_command(client, ["redun", "tag", "list", "step=calc"])
    assert "test.task1 (step=calc)" in output

    output = run_command(client, ["redun", "tag", "list", "step=unknown"])
    assert "test.task1" not in output

    # Add a tag.
    output = run_command(client, ["redun", "tag", "add", execution.id, "env=prod"])
    output = run_command(client, ["redun", "tag", "list", "env=prod"])
    assert "Exec " in output

    # Remove a tag.
    output = run_command(client, ["redun", "tag", "rm", execution.id, "env=prod"])
    output = run_command(client, ["redun", "tag", "list", "env=prod"])
    assert "Exec " not in output

    # Add a tag back.
    output = run_command(client, ["redun", "tag", "add", execution.id, "env=prod"])
    output = run_command(client, ["redun", "tag", "list", "env=prod"])
    assert "Exec " in output

    # Update a tag.
    output = run_command(client, ["redun", "tag", "update", execution.id, "env=prod2"])
    output = run_command(client, ["redun", "tag", "list", "env=prod2"])
    assert "Exec " in output


@use_tempdir
def test_cli_log_tags_order() -> None:
    """
    redun log entries should have tags sorted in a stable way, even if their values are not
    comparable directly.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task
from redun.scheduler import apply_tags

redun_namespace = 'test'

@task()
def example1():
    return apply_tags(
        42,
        execution_tags=[("key", {"b": 1}), ("key", {"a": 2})]
    )

@task()
def example2():
    return apply_tags(
        42,
        execution_tags=[("key", {"c": 2}), ("key", {"c": 1})]
    )

"""
    )

    # Run example workflow.
    client = RedunClient()
    client.execute(["redun", "run", "workflow.py", "example1"])
    client.execute(["redun", "run", "workflow.py", "example2"])

    lines = run_command(client, ["redun", "log"]).split("\n")

    assert_match_lines(
        [
            r"Recent executions:",
            r"",
            r"Exec.*key={\"c\": 1}, key={\"c\": 2}.*",
            r"Exec.*key={\"a\": 2}, key={\"b\": 1}.*",
            r"",
        ],
        lines,
    )


@use_tempdir
def test_cli_log_files() -> None:
    """
    cli should support querying for files.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task, File

@task()
def task1(filename: str):
    file = File(filename)
    file.write('hello')
    return file

@task()
def main(filename: str):
    return task1(filename)
"""
    )

    # Run example workflow.
    client = RedunClient()
    client.execute(["redun", "run", "workflow.py", "main", "--filename", "out"])
    client.execute(["redun", "run", "workflow.py", "main", "--filename", "out2"])
    client.execute(["redun", "run", "workflow.py", "main", "--filename", "out"])

    # Query File type.
    lines = run_command(client, ["redun", "log", "--file"]).split("\n")
    assert_match_lines(
        [
            "File ........ out ",
            "File ........ out2 ",
            "",
        ],
        lines,
    )

    # Query Files by path.
    lines = run_command(client, ["redun", "log", "--file-path", "out"]).split("\n")
    assert_match_lines(
        [r"Value ........ redun.File File\(path=out, hash=........\) ", ""],
        lines,
    )

    # Query file provenance.
    lines = run_command(client, ["redun", "log", "out"]).split("\n")
    assert_match_lines(
        [
            r"File ........ out ",
            r"Produced by Job ........",
            r"",
            r"  Job .* \[ DONE \] ....-..-.. ..:..:..:  task1\('out'\) ",
            r"  Traceback: Exec ........ > Job ........ main > Job ........ task1",
            r"  Duration: 0:00:..\...",
            r"",
            r"    CallNode .* task1 ",
            r"      Args:   'out'",
            r"      Result: File\(path=out, hash=........\)",
            r"",
            r"    Task 8106568b9069e7b743d827a7167bec59bb86d78a task1 ",
            r"",
            r"      def task1\(filename: str\):",
            r"          file = File\(filename\)",
            r"          file.write\('hello'\)",
            r"          return file",
            r"",
            r"",
            r"    Upstream dataflow:",
            r"",
            r"      result = File\(path=out, hash=........\)",
            r"",
            r"      result <-- <........> task1\(filename\)",
            r"        filename = <........> 'out'",
            r"",
            r"      filename <-- argument of <........> main\(filename\)",
            r"               <-- origin",
            r"",
            r"",
        ],
        lines,
    )


def test_port_usage_check() -> None:
    taken_port = 65500
    open_port = 65501

    # Ensure that bound ports are detected as in use
    with socket() as sock1:
        sock1.bind(("localhost", taken_port))
        assert is_port_in_use("localhost", taken_port)
        assert is_port_in_use("localhost", str(taken_port))

    # Ensure that open ports are detected
    assert not is_port_in_use("localhost", open_port)
    # Ensure that the detection process itself doesn't bind the port
    assert not is_port_in_use("localhost", open_port)

    with pytest.raises(ValueError):
        is_port_in_use("localhost", "bla")


@use_tempdir
def test_cli_migrate() -> None:
    """
    cli should work with database migrations.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def main() -> int:
    return 10
"""
    )

    # Run example workflow.
    client = RedunClient()
    assert client.execute(["redun", "run", "workflow.py", "main"]) == 10
    backend = RedunBackendDb(db_uri="sqlite:///.redun/redun.db")
    backend.create_engine()

    # New database should be current.
    assert backend.is_db_compatible()

    # Should be able to get db info.
    lines = run_command(client, ["redun", "db", "info"]).split("\n")
    assert_match_lines(
        [
            r"redun :: version .*",
            r"config dir: .*",
            r"",
            r"db version: \d+(\.\d+)? '.*'",
            r"CLI requires db versions: >=[0-9.]+,<\d+",
            r"CLI compatible with db: True",
            r"",
        ],
        lines,
    )

    # Should be able to downgrade the database.
    client = RedunClient()
    client.execute(["redun", "db", "downgrade", "1"])

    # Database should now be incompatible.
    client = RedunClient()
    client.execute(["redun", "db", "info"])
    backend = RedunBackendDb(db_uri="sqlite:///.redun/redun.db")
    backend.create_engine()
    assert not backend.is_db_compatible()

    # redun log should be prevented.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "log"])

    # By default redun run automigrates.
    client = RedunClient()
    assert client.execute(["redun", "run", "workflow.py", "main"]) == 10
    backend = RedunBackendDb(db_uri="sqlite:///.redun/redun.db")
    backend.create_engine()

    # database should have automigrated.
    assert backend.is_db_compatible()


@use_tempdir
def test_cli_automigrate_postgres() -> None:
    """
    Postgres databases should not automigrate.
    """
    File(".redun2/redun.ini").write(
        """
[backend]
db_uri = postgresql://host/db
"""
    )

    client = RedunClient()
    parser = client.get_command_parser()
    args = parser.parse_args(["-c", ".redun2", "run", "workflow.py", "main"])

    with patch.object(Scheduler, "load") as load_mock:
        client.get_scheduler(args, migrate_if_local=True)

    load_mock.assert_called_with(migrate=False)


@use_tempdir
def test_rerun() -> None:
    """
    cli should allow re-running of subgraphs
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task

redun_namespace = 'test'

@task()
def task1(x: int):
    return x + 1

@task()
def main(x: int, y: int):
    return task1(x) + 1 + y
"""
    )

    client = RedunClient()

    result = client.execute(["redun", "run", "workflow.py", "main", "--x", "10", "--y", "5"])
    assert result == 17

    assert client.scheduler
    scheduler = client.scheduler
    assert scheduler.backend
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.session

    # Ensure that we can rerun from multiple jobs in the call graph.
    jobs = backend.session.query(Job).all()
    assert len(jobs) == 2
    for job in jobs:
        result = client.execute(["redun", "run", "--rerun", "workflow.py", job.id])
        if job.parent_id:
            assert result == 11
        else:
            assert result == 17

    # Ensure that we can specify execution ids as well as job ids
    execution = backend.session.query(Execution).first()
    result = client.execute(["redun", "run", "--rerun", "workflow.py", execution.id])
    assert result == 17

    result = client.execute(["redun", "run", "--rerun", "workflow.py", execution.id, "--y", "10"])
    assert result == 22


@use_tempdir
def test_run_packed_inputs() -> None:
    """
    Test that `run` accepts binary inputs w/ the --input flag.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task

redun_namespace = 'test'

@task()
def main(x: int, y: int = 0):
    # Note: the default value y=0 is important to the test, this is a corner case of the CLI.
    return x + y
"""
    )

    client = RedunClient()

    with File("inputs").open("wb") as out:
        pickle_dump([(5,), {"y": 7}], out)

    result = client.execute(["redun", "run", "workflow.py", "main", "--input", "inputs"])
    assert result == 12

    # We should still allow overrides from the cmdline.
    result = client.execute(
        ["redun", "run", "workflow.py", "main", "--input", "inputs", "--y", "10"]
    )
    assert result == 15


@pytest.mark.parametrize("executor", ["proc", "thread"])
@use_tempdir
def test_launch(executor: str) -> None:
    """
    Launch subcommand should run a workflow within an executor.
    """
    config_string = """
[executors.proc]
type = local
mode = process

[executors.thread]
type = local
mode = thread
"""
    workflow = """
import os
import redun

@redun.task
def main() -> int:
    return os.getpid()
"""
    File(".redun/redun.ini").write(config_string)
    File("workflow.py").write(workflow)
    this_pid = os.getpid()
    client = RedunClient()
    client.execute(
        [
            "redun",
            "launch",
            "--executor",
            executor,
            "redun",
            "run",
            "workflow.py",
            "main",
        ]
    )

    if (sys.version_info.major, sys.version_info.minor) == (
        3,
        8,
    ) and executor == "proc":
        # Process executors are not shut down in py3.8, so it may still be running.
        # This is expected but we skip the test rather than try to wait it out
        return

    parser = client.get_command_parser()
    args, extra_args = parser.parse_known_args([])
    value = client.get_session(args=args).query(Value).filter(Value.type != "redun.Task").one()
    assert value.value_parsed
    assert value.value_parsed != this_pid


@pytest.fixture(scope="module")
def context_workflow():
    body = """
import os
import redun

@redun.task
def main(context: dict = redun.get_context("my_tool", None)) -> dict:
    return context
"""
    return File("workflow.py").write(body)


@use_tempdir
def test_both_context_and_context_file_present(context_workflow) -> None:
    """
    Ensure that an error is raised if both context and context file are present.
    """
    config_string = """
[scheduler]
context_file = context.json
context = {}
"""
    File(".redun/redun.ini").write(config_string)
    client = RedunClient()
    with pytest.raises(SchedulerError, match="Both context and context_file specified in config!"):
        client.execute(["redun", "run", "workflow.py", "main"])


@use_tempdir
def test_context_file_in_redun_config(context_workflow) -> None:
    """
    Ensure that context_file option can be used in redun config.
    """
    config_string = """
[scheduler]
context_file = context.json
"""
    File(".redun/redun.ini").write(config_string)
    File(".redun/context.json").write(
        """{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    )
    client = RedunClient()
    result = client.execute(["redun", "run", "workflow.py", "main"])
    assert result == {"a": 2, "c": "bar"}


@use_tempdir
def test_context_file_override(context_workflow) -> None:
    """
    Ensure that the context from redun.ini can be overridden by the context file.
    """
    config_string = """
[scheduler]
context =
  {
    "my_tool": {
      "a": 1,
      "b": "foo"
    }
  }
"""
    context_override = """
{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    File(".redun/redun.ini").write(config_string)
    File("context.json").write(context_override)
    client = RedunClient()
    result = client.execute(
        ["redun", "run", "--context-file", "context.json", "workflow.py", "main"]
    )
    assert result == {"a": 2, "b": "foo", "c": "bar"}


@use_tempdir
def test_context_file_config_relative(context_workflow) -> None:
    """
    Ensure that when the context file is a relative path specified in redun.ini, it is searched in
    the directory relative to the .redun config dir.
    """
    config_string = """
[scheduler]
context_file = ../context.json
"""
    File(".redun/redun.ini").write(config_string)
    File("context.json").write(
        """{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    )
    client = RedunClient()
    result = client.execute(["redun", "run", "workflow.py", "main"])
    assert result == {"a": 2, "c": "bar"}


@use_tempdir
def test_context_file_config_absolute(context_workflow) -> None:
    """
    Ensure that when the context file is an absolute path specified in redun.ini, it is used as is.
    """
    context_file_path = f"{os.getcwd()}/context.json"
    assert os.path.isabs(context_file_path)

    config_string = f"""
[scheduler]
context_file = {context_file_path}
"""
    File(".redun/redun.ini").write(config_string)
    File(context_file_path).write(
        """{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    )
    client = RedunClient()
    result = client.execute(["redun", "run", "workflow.py", "main"])
    assert result == {"a": 2, "c": "bar"}


@use_tempdir
def test_context_file_arg_relative(context_workflow) -> None:
    """
    Ensure that when the context file is a relative path specified as --context-file argument,
    it is searched in the current working directory.
    """
    File("context.json").write(
        """{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    )
    client = RedunClient()
    result = client.execute(
        ["redun", "run", "--context-file", "context.json", "workflow.py", "main"]
    )
    assert result == {"a": 2, "c": "bar"}


@use_tempdir
def test_context_file_arg_absolute(context_workflow) -> None:
    """
    Ensure that when the context file is an absolute path specified as --context-file argument,
    it is used as is.
    """
    context_file_path = f"{os.getcwd()}/context.json"
    assert os.path.isabs(context_file_path)

    File(context_file_path).write(
        """{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    )
    client = RedunClient()
    result = client.execute(
        ["redun", "run", "--context-file", context_file_path, "workflow.py", "main"]
    )
    assert result == {"a": 2, "c": "bar"}


@use_tempdir
def test_context_option_override(context_workflow) -> None:
    """
    Ensure that the context from redun.ini can be overridden by the context option.
    """
    config_string = """
[scheduler]
context =
  {
    "my_tool": {
      "a": 1,
      "b": "foo"
    }
  }
"""
    File(".redun/redun.ini").write(config_string)
    client = RedunClient()
    result = client.execute(
        [
            "redun",
            "run",
            "--context",
            "my_tool.a=2",
            "--context",
            "my_tool.c=bar",
            "workflow.py",
            "main",
        ]
    )
    assert result == {"a": 2, "b": "foo", "c": "bar"}


@use_tempdir
def test_context_file_and_option_override(context_workflow) -> None:
    """
    Ensure that the context from redun.ini can be overridden by the context file and option.
    Overrides from --context should take precedence over the context file and the context file
    should take precedence over the redun.ini context.
    """
    config_string = """
[scheduler]
context =
  {
    "my_tool": {
      "a": 1,
      "b": "foo"
    }
  }
"""
    context_override = """
{
    "my_tool": {
        "a": 2,
        "c": "bar"
    }
}
"""
    File(".redun/redun.ini").write(config_string)
    File("context.json").write(context_override)
    client = RedunClient()
    result = client.execute(
        [
            "redun",
            "run",
            "--context-file",
            "context.json",
            "--context",
            "my_tool.a=3",
            "--context",
            "my_tool.d=baz",
            "workflow.py",
            "main",
        ]
    )
    assert result == {"a": 3, "b": "foo", "c": "bar", "d": "baz"}


@use_tempdir
def test_context_update_context_precedence() -> None:
    """
    Ensure that update_context takes precedence over any other option.
    """
    config_string = """
[scheduler]
context =
  {
    "my_tool": {
      "a": 1,
      "b": "foo"
    }
  }
"""
    workflow = """
import os
import redun

@redun.task
def inner(context: dict = redun.get_context("my_tool", None)) -> dict:
    return context

@redun.task
def main() -> dict:
    return inner.update_context(my_tool={"a": 3, "c": "bar"})()
"""
    File(".redun/redun.ini").write(config_string)
    File("workflow.py").write(workflow)
    client = RedunClient()
    result = client.execute(
        [
            "redun",
            "run",
            "--context",
            "my_tool.a=2",
            "workflow.py",
            "main",
        ]
    )
    assert result == {"a": 3, "b": "foo", "c": "bar"}


@pytest.mark.parametrize("executor", ["proc", "thread"])
@use_tempdir
def test_wait_launch(executor: str) -> None:
    """
    Launch with --wait which prints the value
    """
    config_string = """
[executors.proc]
type = local
mode = process

[executors.thread]
type = local
mode = thread
"""
    workflow = """
import os
import redun

@redun.task
def main() -> str:
    return "hi"
"""
    File(".redun/redun.ini").write(config_string)
    File("workflow.py").write(workflow)
    client = RedunClient()
    waited_display = run_command(
        client,
        [
            "redun",
            "launch",
            "--wait",
            "--executor",
            executor,
            "redun",
            "run",
            "workflow.py",
            "main",
        ],
    )
    lines = waited_display.split("\n")
    assert lines[-2] == "b\"'hi'\\n\""



import boto3

from redun import File, task
from redun.executors.command import get_oneshot_command, get_script_task_command
from redun.scheduler import Job
from redun.tests.utils import mock_s3, use_tempdir


@task
def task1(x):
    return x


@use_tempdir
def test_get_oneshot_command() -> None:
    """
    Should be able to generate a shell command for redun oneshot.
    """

    expr = task1(10)
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"
    code_file = File("scratch/code.tar.gz")

    command = get_oneshot_command(
        "scratch",
        job,
        task1,
        (10,),
        {},
        {},
        code_file=code_file,
    )
    assert command == [
        "redun",
        "--check-version",
        ">=0.4.1",
        "oneshot",
        "redun.tests.test_executor_command",
        "--code",
        "scratch/code.tar.gz",
        "--input",
        "scratch/jobs/eval_hash/input",
        "--output",
        "scratch/jobs/eval_hash/output",
        "--error",
        "scratch/jobs/eval_hash/error",
        "task1",
    ]


@use_tempdir
def test_get_script_task_command_local() -> None:
    """
    Should be able to generate a shell command for a redun script task.
    """
    expr = task1(10)
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"

    command = get_script_task_command("scratch", job, "myprog -x 1 input.txt")
    assert command == [
        "bash",
        "-c",
        "-o",
        "pipefail",
        "\n"
        "cp scratch/jobs/eval_hash/input .task_command\n"
        "chmod +x .task_command\n"
        "(\n"
        "  ./.task_command   2> >(tee .task_error >&2) | tee .task_output\n"
        ") && (\n"
        "    cp .task_output scratch/jobs/eval_hash/output\n"
        "    cp .task_error scratch/jobs/eval_hash/error\n"
        "    echo ok | cat > scratch/jobs/eval_hash/status\n"
        ") || (\n"
        "    [ -f .task_output ] && cp .task_output scratch/jobs/eval_hash/output\n"
        "    [ -f .task_error ] && cp .task_error scratch/jobs/eval_hash/error\n"
        "    echo fail | cat > scratch/jobs/eval_hash/status\n"
        "    \n"
        ")\n",
    ]


@use_tempdir
@mock_s3
def test_get_script_task_command_s3() -> None:
    """
    Should be able to generate a shell command for a redun script task.
    """
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="bucket")

    expr = task1(10)
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"

    command = get_script_task_command("s3://bucket/scratch", job, "myprog -x 1 input.txt")
    assert command == [
        "bash",
        "-c",
        "-o",
        "pipefail",
        "\n"
        "aws s3 cp --no-progress s3://bucket/scratch/jobs/eval_hash/input .task_command\n"
        "chmod +x .task_command\n"
        "(\n"
        "  ./.task_command   2> >(tee .task_error >&2) | tee .task_output\n"
        ") && (\n"
        "    aws s3 cp --no-progress .task_output s3://bucket/scratch/jobs/eval_hash/output\n"
        "    aws s3 cp --no-progress .task_error s3://bucket/scratch/jobs/eval_hash/error\n"
        "    echo ok | aws s3 cp --no-progress - s3://bucket/scratch/jobs/eval_hash/status\n"
        ") || (\n"
        "    [ -f .task_output ] && aws s3 cp --no-progress .task_output "
        "s3://bucket/scratch/jobs/eval_hash/output\n"
        "    [ -f .task_error ] && aws s3 cp --no-progress .task_error "
        "s3://bucket/scratch/jobs/eval_hash/error\n"
        "    echo fail | aws s3 cp --no-progress - s3://bucket/scratch/jobs/eval_hash/status\n"
        "    \n"
        ")\n",
    ]



"""add task version and backfill companion values

Revision ID: 30ffbaee18cd
Revises: 647c510a77b1
Create Date: 2021-05-19 16:55:43.724884

"""

from typing import Iterator

from alembic import context, op
from sqlalchemy.orm import Session

from redun.backends import db
from redun.task import Task
from redun.utils import pickle_dumps
from redun.value import MIME_TYPE_PICKLE

# revision identifiers, used by Alembic.
revision = "30ffbaee18cd"
down_revision = "647c510a77b1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # ### end Alembic commands ###
    # Data migration

    # Do not try to do data migration in offline mode.
    if context.is_offline_mode():
        return

    session = Session(bind=op.get_bind())
    try:
        backfill_values_for_lonely_tasks(session)
    finally:
        session.close()


def downgrade():
    # Deliberate choice to not roll back `backfill_values_for_lonely_tasks`
    # ### commands auto generated by Alembic - please adjust! ###
    # ### end Alembic commands ###
    pass


# Data migration helpers
def backfill_values_for_lonely_tasks(session: Session) -> None:
    """
    Create a Value record corresponding to every Task in the database, if it doesn't already exist.
    """
    lonely_db_tasks = get_lonely_tasks(session)
    for db_task in lonely_db_tasks:
        # We can't get `func` from the registry, so we use a dummy function when instantiating
        # the `redun.Task` corresponding to the `db_task`. This is acceptable for our needs since
        # func is not consulted during serialization when other parameters (like version) are
        # passed as arguments.
        redun_task: Task = Task(
            func=lambda: None,
            name=db_task.name,
            namespace=db_task.namespace,
            compat=[],  # Missing from db record
            script=guess_is_script(db_task),
            task_options_base={},  # Missing from task record & not hashed until very recently
        )
        session.add(
            db.Value(
                value_hash=db_task.hash,
                type=Task.type_name,
                format=MIME_TYPE_PICKLE,
                value=pickle_dumps(redun_task),
            )
        )
    if session.new:
        session.commit()


def get_lonely_tasks(session: Session) -> Iterator[db.Task]:
    """
    Query the database for all Tasks that don't have corresponding Value records
    """
    return session.query(db.Task).filter_by(value=None).all()


def guess_is_script(task: db.Task) -> bool:
    """
    Check whether an input task is among those known to be script tasks. This is a best-guess
    based on perusing existing redun usage in the insitro GitHub organization, so it's likely that
    there will be some false-negatives.
    """
    known_script_tasks = {
        "redun.script_task",
        "test_help",
        "test_help_debug",
        "test_gfetch",
    }
    return task.fullname in known_script_tasks



import abc
import datetime
from base64 import b64decode, b64encode
from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional

from dateutil.parser import parse as parse_date
from sqlalchemy.orm import aliased, joinedload, selectinload
from sqlalchemy.orm.query import Query

from redun.backends import db
from redun.backends.base import TagEntity

# Serialization version.
VERSION = 1


def serialize_timestamp(timestamp: Optional[datetime.datetime]) -> Optional[str]:
    """
    Serialize a datetime value to a timestamp string.
    """
    if timestamp:
        return str(timestamp)
    else:
        return None


def deserialize_timestamp(date_str: Optional[str]) -> Optional[datetime.datetime]:
    """
    Deserialize timestamp string to datetime value.
    """
    if date_str:
        return parse_date(date_str)
    else:
        return None


class Serializer(abc.ABC):
    """
    Base class serializer for RedunBackendDb records.
    """

    pk_field: str = ""

    def __init__(self, db_version: "Optional[db.DBVersionInfo]" = None):
        self._db_version = db_version or db.REDUN_DB_VERSIONS[-1]

    def get_pk(self, spec: dict) -> str:
        return spec[self.pk_field]

    def serialize(self, row: "db.Base") -> dict:
        raise NotImplementedError()

    def serialize_query(self, query: Query) -> Iterator[dict]:
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        raise NotImplementedError()


class ExecutionSerializer(Serializer):
    pk_field = "id"

    def serialize(self, execution: "db.Base") -> dict:
        return {
            "_version": VERSION,
            "_type": "Execution",
            "id": execution.id,
            "args": execution.args,
            "job_id": execution.job_id,
        }

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Execution"
        return [
            db.Execution(
                id=spec["id"],
                args=spec["args"],
                job_id=spec["job_id"],
            )
        ]


class JobSerializer(Serializer):
    pk_field = "id"

    def serialize(self, job: "db.Base") -> dict:
        return {
            "_version": VERSION,
            "_type": "Job",
            "id": job.id,
            "start_time": serialize_timestamp(job.start_time),
            "end_time": serialize_timestamp(job.end_time),
            "task_hash": job.task_hash,
            "cached": job.cached,
            "status": job.status,
            "call_hash": job.call_hash,
            "parent_id": job.parent_id,
            "children": [child.id for child in job.child_jobs],
            "execution_id": job.execution_id,
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        # Fetch parent-child links.
        Job = aliased(db.Job, query.subquery())
        ChildJob = aliased(db.Job, query.subquery())

        parent_child_pairs = (
            query.session.query(Job.id, ChildJob.id)
            .join(ChildJob, ChildJob.parent_id == Job.id)
            .order_by(ChildJob.start_time)
            .all()
        )
        parent2children = defaultdict(list)
        for parent_id, child_id in parent_child_pairs:
            parent2children[parent_id].append(child_id)

        # Fetch status in bulk.
        Value = aliased(db.Value)
        job_result_types = (
            query.session.query(Job.id, Value.type)
            .outerjoin(db.CallNode, Job.call_hash == db.CallNode.call_hash)
            .outerjoin(Value, db.CallNode.value_hash == Value.value_hash)
        ).all()
        job_id2result_type = {job_id: result_type for job_id, result_type in job_result_types}

        for job in query.all():
            yield {
                "_version": VERSION,
                "_type": "Job",
                "id": job.id,
                "start_time": serialize_timestamp(job.start_time),
                "end_time": serialize_timestamp(job.end_time),
                "task_hash": job.task_hash,
                "cached": job.cached,
                "status": job.calc_status(job_id2result_type[job.id]),
                "call_hash": job.call_hash,
                "parent_id": job.parent_id,
                "children": parent2children[job.id],
                "execution_id": job.execution_id,
            }

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Job"
        return [
            db.Job(
                id=spec["id"],
                start_time=deserialize_timestamp(spec["start_time"]),
                end_time=deserialize_timestamp(spec["end_time"]),
                task_hash=spec["task_hash"],
                cached=spec["cached"],
                call_hash=spec["call_hash"],
                parent_id=spec["parent_id"],
                execution_id=spec["execution_id"],
            )
        ]


class ValueSerializer(Serializer):
    pk_field = "value_hash"

    def serialize(self, value: "db.Base") -> dict:
        # Up to one special subtype should exist at a time
        assert not (bool(value.file) and bool(value.task))
        subtype: Optional[Dict[str, Any]] = None
        if value.file:
            subtype = {
                "_type": "File",
                "path": value.file.path,
            }
        elif value.task:
            subtype = {
                "_type": "Task",
                "name": value.task.name,
                "namespace": value.task.namespace,
                "source": value.task.source,
            }
        return {
            "_version": VERSION,
            "_type": "Value",
            "value_hash": value.value_hash,
            "type": value.type,
            "format": value.format,
            "data": b64encode(value.value).decode("utf8"),
            "subvalues": sorted(edge.value_hash for edge in value.child_edges),
            "subtype": subtype,
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        query = query.options(
            joinedload(db.Value.file),
            joinedload(db.Value.task),
            joinedload(db.Value.child_edges),
        )
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Value"
        return (
            [
                db.Value(
                    value_hash=spec["value_hash"],
                    type=spec["type"],
                    format=spec["format"],
                    value=b64decode(spec["data"]),
                ),
            ]
            + [
                db.Subvalue(parent_value_hash=spec["value_hash"], value_hash=child_id)
                for child_id in spec["subvalues"]
            ]
            + (
                [
                    db.File(
                        value_hash=spec["value_hash"],
                        path=spec["subtype"]["path"],
                    )
                ]
                if spec.get("subtype") and spec["subtype"]["_type"] == "File"
                else []
            )
            + (
                [
                    db.Task(
                        hash=spec["value_hash"],
                        name=spec["subtype"]["name"],
                        namespace=spec["subtype"]["namespace"],
                        source=spec["subtype"]["source"],
                    )
                ]
                if spec.get("subtype") and spec["subtype"]["_type"] == "Task"
                else []
            )
        )


class CallNodeSerializer(Serializer):
    pk_field = "call_hash"

    def serialize(self, call_node: "db.Base") -> dict:
        args = {
            arg.arg_key or str(arg.arg_position): {
                "arg_hash": arg.arg_hash,
                "value_hash": arg.value_hash,
                "upstream": sorted(arg_result.result_call_hash for arg_result in arg.arg_results),
            }
            for arg in call_node.arguments
        }

        return {
            "_version": VERSION,
            "_type": "CallNode",
            "call_hash": call_node.call_hash,
            "task_name": call_node.task_name,
            "task_hash": call_node.task_hash,
            "args_hash": call_node.args_hash,
            "value_hash": call_node.value_hash,
            "timestamp": serialize_timestamp(call_node.timestamp),
            "args": args,
            "children": [edge.child_id for edge in call_node.child_edges],
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        query = query.options(
            selectinload(db.CallNode.arguments).joinedload(db.Argument.arg_results),
            selectinload(db.CallNode.child_edges),
        )
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "CallNode"

        return (
            [
                db.CallNode(
                    call_hash=spec["call_hash"],
                    task_name=spec["task_name"],
                    task_hash=spec["task_hash"],
                    args_hash=spec["args_hash"],
                    value_hash=spec["value_hash"],
                    timestamp=deserialize_timestamp(spec["timestamp"]),
                )
            ]
            + [
                db.CallEdge(
                    parent_id=spec["call_hash"],
                    child_id=child_hash,
                    call_order=call_index,
                )
                for call_index, child_hash in enumerate(spec["children"])
            ]
            + [
                db.Argument(
                    arg_hash=arg["arg_hash"],
                    call_hash=spec["call_hash"],
                    value_hash=arg["value_hash"],
                    arg_position=int(key) if key.isdigit() else None,
                    arg_key=key if not key.isdigit() else None,
                )
                for key, arg in spec["args"].items()
            ]
            + [
                db.ArgumentResult(
                    arg_hash=arg["arg_hash"],
                    result_call_hash=upstream,
                )
                for key, arg in spec["args"].items()
                for upstream in arg["upstream"]
            ]
        )


class TagSerializer(Serializer):
    pk_field = "tag_hash"

    def serialize(self, tag: "db.Base") -> dict:
        return {
            "_version": VERSION,
            "_type": "Tag",
            "tag_hash": tag.tag_hash,
            "entity_type": tag.entity_type.name,
            "entity_id": tag.entity_id,
            "key": tag.key,
            "value": tag.value,
            "parents": [edit.parent_id for edit in tag.parent_edits],
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        query = query.options(
            selectinload(db.Tag.parent_edits).joinedload(db.TagEdit.child),
        )
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Tag"

        return [
            db.Tag(
                tag_hash=spec["tag_hash"],
                entity_type=TagEntity(spec["entity_type"]),
                entity_id=spec["entity_id"],
                key=spec["key"],
                value=spec["value"],
            )
        ] + [db.TagEdit(parent_id=parent, child_id=spec["tag_hash"]) for parent in spec["parents"]]


class RecordSerializer(Serializer):
    def __init__(self, db_version: "Optional[db.DBVersionInfo]" = None):
        super().__init__(db_version)
        self._serializers = {
            "Execution": ExecutionSerializer(db_version),
            "Job": JobSerializer(db_version),
            "CallNode": CallNodeSerializer(db_version),
            "Value": ValueSerializer(db_version),
            "Tag": TagSerializer(db_version),
        }

    def get_pk(self, spec: dict) -> str:
        return self._serializers[spec["_type"]].get_pk(spec)

    def get_subserializer(self, type: str) -> Serializer:
        return self._serializers[type]

    def serialize(self, obj: "db.Base") -> dict:
        if isinstance(obj, db.Task):
            # Use Value-version of Task for serialization.
            obj = obj.value
        return self._serializers[type(obj).__name__].serialize(obj)

    def serialize_query(self, query: Query) -> Iterator[dict]:
        [column] = query.column_descriptions
        return self._serializers[column["name"]].serialize_query(query)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        return self._serializers[spec["_type"]].deserialize(spec)



import os
from configparser import SectionProxy
from unittest.mock import patch

import pytest

from redun.cli import get_config_dir, setup_scheduler
from redun.config import Config
from redun.file import RedunFileNotFoundError
from redun.scheduler_config import postprocess_config
from redun.tests.utils import use_tempdir


def test_config_parse_sections():
    config_string = """
[executors.batch]
type = aws_batch
image = 123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3
queue = queue
s3_scratch = s3://example-bucket/redun/

[executors.batch2]
type = aws_batch
image = 123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3
queue = queue2
s3_scratch = s3://example-bucket/redun/
job_def_extra = {"a_json_key": {"a_json_nested_key": 42}}
"""

    config = Config()
    config.read_string(config_string)

    assert sorted(config.keys()) == ["executors"]
    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["batch"]["queue"] == "queue"
    assert config["executors"]["batch2"]["type"] == "aws_batch"
    assert config["executors"]["batch2"]["queue"] == "queue2"
    # ensure json property is loaded as text
    assert config["executors"]["batch2"]["job_def_extra"] == (
        '{"a_json_key": {"a_json_nested_key": 42}}'
    )


@pytest.fixture
def sample_config():
    return Config(
        {
            "executors.batch": {
                "type": "aws_batch",
                "image": "123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3",
                "queue": "queue",
                "s3_scratch": "s3://example-bucket/redun/",
            },
            "executors.batch2": {
                "type": "aws_batch",
                "image": "123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3",
                "queue": "queue2",
                "s3_scratch": "s3://example-bucket/redun/",
            },
        }
    )


def test_config_dict(sample_config):
    config = sample_config

    assert sorted(config.keys()) == ["executors"]
    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["batch"]["queue"] == "queue"
    assert config["executors"]["batch2"]["type"] == "aws_batch"
    assert config["executors"]["batch2"]["queue"] == "queue2"


def test_get_config_dict(sample_config):
    """Assert get_config_dict() reproduces a Config correctly"""
    config = Config(config_dict=sample_config.get_config_dict())
    assert sorted(config.keys()) == ["executors"]
    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["batch"]["queue"] == "queue"
    assert config["executors"]["batch2"]["type"] == "aws_batch"
    assert config["executors"]["batch2"]["queue"] == "queue2"


@use_tempdir
def test_get_config_dict_replace_config_dir(sample_config):
    """Verify replace_config_dir behavior"""
    os.makedirs(".redun")
    with open(".redun/redun.ini", "w") as out:
        out.write(
            """
[backend]
db_uri = sqlite:///redun.db
"""
        )

    scheduler = setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists(".redun/redun.ini")
    assert os.path.exists(".redun/redun.db")

    conf = scheduler.config.get_config_dict()
    local_config_dir = get_config_dir()
    db_uri = conf["backend"]["db_uri"]
    assert db_uri == f"sqlite:///{local_config_dir}/redun.db"
    assert conf["backend"]["config_dir"] == local_config_dir
    assert conf["repos.default"]["config_dir"] == local_config_dir

    replacement = "/tmp/xyz"
    conf = scheduler.config.get_config_dict(replace_config_dir=replacement)
    db_uri = conf["backend"]["db_uri"]
    assert db_uri == f"sqlite:///{replacement}/redun.db"
    assert conf["backend"]["config_dir"] == replacement
    assert conf["repos.default"]["config_dir"] == replacement


def test_postprocess_config(sample_config):
    """
    Default sections must be true Sections.
    If they aren't an exception will occur when get_config_dict() is called
    """
    post_processed_config = postprocess_config(sample_config, "/tmp/config_dir")
    config = Config(config_dict=post_processed_config.get_config_dict())

    assert sorted(config.keys()) == ["backend", "executors", "repos", "scheduler"]
    config = postprocess_config(config, "/tmp/config_dir")
    # Default repository should be added since it was missing
    assert config["repos"]["default"]["config_dir"] == "/tmp/config_dir"
    # db_uri should be converted to absolute absolute config_dir
    assert config["backend"]["db_uri"] == "sqlite:////tmp/config_dir/redun.db"

    # Ensure that the auto-added sections are truly Section objects and not plain dicts.
    assert isinstance(config["repos"]["default"], SectionProxy)
    assert isinstance(config["backend"], SectionProxy)


def test_postprocess_federated_config(sample_config):
    federated_imports_dict = sample_config.get_config_dict()
    federated_imports_dict["scheduler"] = {
        "federated_configs": """redun/tests/test_data/federated_configs/federated_import_valid"""
    }
    config = Config(federated_imports_dict)
    config = postprocess_config(config, "/tmp/config_dir")

    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["headnode"]["type"] == "docker"
    assert config["federated_tasks"]["entrypoint_name"]["executor"] == "headnode"

    federated_imports_dict["scheduler"] = {
        "federated_configs": """redun/tests/test_data/federated_configs/federated_import_valid
redun/tests/test_data/federated_configs/federated_import_valid2"""
    }

    config = Config(federated_imports_dict)
    config = postprocess_config(config, "/tmp/config_dir")

    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["headnode"]["type"] == "docker"
    assert config["executors"]["headnode2"]["type"] == "docker"
    assert config["federated_tasks"]["entrypoint_name"]["executor"] == "headnode"
    assert config["federated_tasks"]["entrypoint_name2"]["executor"] == "headnode2"

    # Check duplicate handling on executors against the main config
    federated_imports_dict["executors.headnode"] = {"type": "duplicate"}
    config = Config(federated_imports_dict)

    with pytest.raises(AssertionError, match="Imported executor name `headnode` is a duplicate"):
        postprocess_config(config, "/tmp/config_dir")

    # Check duplicate handling on tasks  against the main config
    del federated_imports_dict["executors.headnode"]
    federated_imports_dict["federated_tasks.entrypoint_name"] = {"executor": "duplicate"}
    config = Config(federated_imports_dict)

    with pytest.raises(
        AssertionError,
        match="Imported federated_task name `entrypoint_name` is a duplicate",
    ):
        postprocess_config(config, "/tmp/config_dir")

    # Check bad config dir
    federated_imports_dict = sample_config.get_config_dict()
    federated_imports_dict["scheduler"] = {
        "federated_configs": """redun/tests/test_data/federated_configs/federated_import_valid
redun/tests/test_data/federated_configs/not_real_directory"""
    }

    config = Config(federated_imports_dict)

    with pytest.raises(
        RedunFileNotFoundError,
        match="No such file or directory. "
        "redun/tests/test_data/federated_configs/not_real_directory/redun.ini",
    ):
        postprocess_config(config, "/tmp/config_dir")

    # Check conflicts between multiple configs
    federated_imports_dict = sample_config.get_config_dict()
    federated_imports_dict["scheduler"] = {
        "federated_configs": """redun/tests/test_data/federated_configs/federated_import_valid
redun/tests/test_data/federated_configs/federated_import_duplicate_task"""
    }
    config = Config(federated_imports_dict)
    with pytest.raises(
        AssertionError,
        match="Imported federated_task name `entrypoint_name` is a duplicate",
    ):
        postprocess_config(config, "/tmp/config_dir")

    federated_imports_dict["scheduler"] = {
        "federated_configs": """redun/tests/test_data/federated_configs/federated_import_valid
redun/tests/test_data/federated_configs/federated_import_duplicate_executor"""
    }

    config = Config(federated_imports_dict)
    with pytest.raises(AssertionError, match="Imported executor name `headnode` is a duplicate"):
        postprocess_config(config, "/tmp/config_dir")


def test_config_env_vars():
    """
    Environment variables should be replaced in config values.
    """
    config_string = """
[DEFAULT]
default_scratch = s3://example-bucket/redun/
queue = default-queue

[executors.batch]
type = aws_batch
image = 123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3
queue = ${QUEUE}
s3_scratch = ${default_scratch}
role = ${ROLE}
"""

    with patch("os.environ", {"ROLE": "my-role", "QUEUE": "my-queue"}):
        config = Config()
        config.read_string(config_string)

        assert config["executors"]["batch"]["role"] == "my-role"

        # We should be able to reference default values.
        assert config["executors"]["batch"]["s3_scratch"] == "s3://example-bucket/redun/"

        # The environment variable should override the default value.
        assert config["executors"]["batch"]["queue"] == "my-queue"


def test_config_case_sensitive():
    """
    Config variables should be case sensitive.
    """
    config_string = """
[section]
key = foo
KEY = bar
"""

    config = Config()
    config.read_string(config_string)

    assert config["section"]["key"] == "foo"
    assert config["section"]["KEY"] == "bar"



"""
From:
    https://github.com/flying-sheep/bcode/blob/master/test.py

Main modification was to remove the __main__ block as we run this as part of our test suite and
not as an isolated test as was done in bcode. A test for the modified encoding behavior was also
added. Formatting and linting changes were also applied.

Original License:

Copyright (c) 2013 Phil Schaf

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from io import BytesIO

from pytest import raises

from redun.bcoding import bdecode, bencode


def test_bools_fail_encoding():
    """
    Bools are recognized as ints by bcode-derived code so we disallow them.
    """
    with raises(TypeError):
        bencode(True)

    # Make sure our modification did not affect the encoding of ints.
    assert bencode(123) == b"i123e"


# basic en/decoding


def test_stream_decoding():
    with BytesIO(b"d2:hii1ee") as f:
        mapping = bdecode(f)
    assert mapping["hi"] == 1


def test_buffer_decoding():
    assert bdecode(b"3:one") == "one"
    assert bdecode("3:two") == "two"


def test_stream_encoding():
    with BytesIO() as stream:
        bencode({"a": 0}, stream)
        assert stream.getvalue() == b"d1:ai0ee"


def test_buffer_encoding():
    assert bencode(("a", 0)) == b"l1:ai0ee"


# decode incomplete stuff


def test_decode_incomplete_int():
    with raises(ValueError):
        print(bdecode("i1"))


def test_decode_incomplete_buffer():
    with raises(ValueError):
        bdecode("1:")


def test_decode_incomplete_list():
    with raises(TypeError):
        print(bdecode("l"))


def test_decode_incomplete_dict():
    with raises(TypeError):
        print(bdecode("d1:k"))



from redun import task


@task
def isolated_task(x):
    """This task is isolated in a module by itself. So, it cannot be used unless the module
    is actually imported."""
    return {
        "result": x,
    }



import json
from unittest.mock import patch

import pytest

from redun.backends.db import RedunBackendDb as RDB
from redun.backends.db import RedunDatabaseError
from redun.config import Section


@pytest.fixture
def mock_env_default_credentials(monkeypatch) -> None:
    monkeypatch.setenv("REDUN_DB_USERNAME", "user")
    monkeypatch.setenv("REDUN_DB_PASSWORD", "secret")


@pytest.fixture
def mock_env_custom_credentials(monkeypatch) -> None:
    monkeypatch.setenv("CUSTOM_DB_USERNAME", "user")
    monkeypatch.setenv("CUSTOM_DB_PASSWORD", "secret")


@pytest.fixture
def mock_env_no_credentials(monkeypatch) -> None:
    monkeypatch.delenv("REDUN_DB_USERNAME", raising=False)
    monkeypatch.delenv("REDUN_DB_PASSWORD", raising=False)


def _check_provided_credentials(conf: Section) -> None:
    sqllite_base = "sqlite:///redun.db"
    assert RDB._get_uri(sqllite_base, conf) == sqllite_base

    postgres_base = "postgresql://localhost:5432/redun"
    postgres_creds = "postgresql://user:secret@localhost:5432/redun"
    assert RDB._get_uri(postgres_base, conf) == postgres_creds


def test_default_credentials(mock_env_default_credentials) -> None:
    _check_provided_credentials({})


def test_custom_credentials(mock_env_custom_credentials) -> None:
    conf = {
        "db_username_env": "CUSTOM_DB_USERNAME",
        "db_password_env": "CUSTOM_DB_PASSWORD",
    }
    _check_provided_credentials(conf)


def test_no_credentials(mock_env_no_credentials) -> None:
    postgres_base = "postgresql://localhost:5432/redun"
    assert RDB._get_uri(postgres_base, {}) == postgres_base


def test_uri_with_credentials() -> None:
    with pytest.raises(RedunDatabaseError):
        RDB._get_uri("postgresql://user:secret@localhost:5432/redun", {})


@patch("redun.executors.aws_utils.get_aws_client")
def test_uri_from_secret(client_mock) -> None:
    engine = "postgres"
    username = "user1"
    password = "secret"
    host = "localhost"
    port = "5555"
    dbname = "mydb"

    client_mock.return_value.get_secret_value.return_value = {
        "SecretString": json.dumps(
            {
                "engine": engine,
                "username": username,
                "password": password,
                "host": host,
                "port": port,
                "dbname": dbname,
            }
        )
    }

    db_uri = RDB._get_uri_from_secret("mysecret")
    assert client_mock.get_secret_value.called_with("mysecret")
    assert db_uri == f"postgresql://{username}:{password}@{host}:{port}/{dbname}"
